import logging
from pathlib import Path

import librosa
import numpy as np
import onnxruntime as ort
import torch

from audiozen.metrics.metric_utils import Metric


options = ort.SessionOptions()
options.intra_op_num_threads = 1
options.inter_op_num_threads = 1

logger = logging.getLogger(__name__)


class pDNSMOS:
    def __init__(self, input_sr=16000) -> None:
        super().__init__()

        root_dir = Path(__file__).parent.absolute()

        self.p835_personal_sess = ort.InferenceSession(
            root_dir / "external" / "pDNSMOS" / "sig_bak_ovr.onnx",
            providers=["CPUExecutionProvider"],
        )

        self.input_sr = input_sr

    def audio_melspec(self, audio, n_mels=120, frame_size=320, hop_length=160, sr=16000, to_db=True):
        mel_spec = librosa.feature.melspectrogram(
            y=audio, sr=sr, n_fft=frame_size + 1, hop_length=hop_length, n_mels=n_mels
        )
        if to_db:
            mel_spec = (librosa.power_to_db(mel_spec, ref=np.max) + 40) / 40  # type: ignore
        return mel_spec.T

    def get_polyfit_val(self, sig, bak, ovr, is_personalized_MOS=False):
        if is_personalized_MOS:
            p_ovr = np.poly1d([-0.00533021, 0.005101, 1.18058466, -0.11236046])
            p_sig = np.poly1d([-0.01019296, 0.02751166, 1.19576786, -0.24348726])
            p_bak = np.poly1d([-0.04976499, 0.44276479, -0.1644611, 0.96883132])
        else:
            p_ovr = np.poly1d([-0.06766283, 1.11546468, 0.04602535])
            p_sig = np.poly1d([-0.08397278, 1.22083953, 0.0052439])
            p_bak = np.poly1d([-0.13166888, 1.60915514, -0.39604546])

        sig_poly = p_sig(sig)
        bak_poly = p_bak(bak)
        ovr_poly = p_ovr(ovr)

        return sig_poly, bak_poly, ovr_poly

    def __call__(self, audio):
        if audio.ndim != 1:
            audio = audio.reshape(-1)

        if torch.is_tensor(audio):
            audio = audio.detach().cpu().numpy()

        SAMPLERATE = 16000
        INPUT_LENGTH = 9.01

        if self.input_sr != 16000:
            audio = librosa.resample(audio, orig_sr=self.input_sr, target_sr=SAMPLERATE)

        len_samples = int(INPUT_LENGTH * SAMPLERATE)

        while len(audio) < len_samples:
            audio = np.append(audio, audio)

        num_hops = int(np.floor(len(audio) / SAMPLERATE) - INPUT_LENGTH) + 1

        hop_len_samples = SAMPLERATE
        predicted_p_mos_sig_seg_raw = []
        predicted_p_mos_bak_seg_raw = []
        predicted_p_mos_ovr_seg_raw = []

        for idx in range(num_hops):
            audio_seg = audio[int(idx * hop_len_samples) : int((idx + INPUT_LENGTH) * hop_len_samples)]
            if len(audio_seg) < len_samples:
                continue

            input_features = np.array(audio_seg).astype("float32")[np.newaxis, :]
            # p808_input_features = np.array(self.audio_melspec(audio=audio_seg[:-160])).astype("float32")[
            # np.newaxis, :, :
            # ]
            oi = {"input_1": input_features}
            # p808_oi = {"input_1": p808_input_features}
            p_mos_sig_raw, p_mos_bak_raw, p_mos_ovr_raw = self.p835_personal_sess.run(None, oi)[0][0]

            predicted_p_mos_ovr_seg_raw.append(p_mos_ovr_raw)
            predicted_p_mos_sig_seg_raw.append(p_mos_sig_raw)
            predicted_p_mos_bak_seg_raw.append(p_mos_bak_raw)

        clip_dict = {}
        # clip_dict["sr"] = SAMPLERATE
        # clip_dict["len"] = actual_audio_len / SAMPLERATE
        clip_dict["pSIG"] = np.mean(predicted_p_mos_sig_seg_raw)
        clip_dict["pBAK"] = np.mean(predicted_p_mos_bak_seg_raw)
        clip_dict["pOVRL"] = np.mean(predicted_p_mos_ovr_seg_raw)

        return clip_dict


class DNSMOS:
    def __init__(self, input_sr=16000, device=-1) -> None:
        super().__init__()

        root_dir = Path(__file__).parent.absolute()

        # =================== Hack to fix the issue with onnxruntime - Start ===================
        # https://github.com/microsoft/onnxruntime/issues/8313#issuecomment-1486097717
        # Globally modify the default inter_op_num_threads and intra_op_num_threads parameters
        # (Only needed for Slurm system as it has a cgroup)
        _default_session_options = ort.capi._pybind_state.get_default_session_options()

        def get_default_session_options_new():
            _default_session_options.inter_op_num_threads = 1
            _default_session_options.intra_op_num_threads = 1
            return _default_session_options

        ort.capi._pybind_state.get_default_session_options = get_default_session_options_new
        # =================== Hack to fix the issue with onnxruntime - End ===================

        if device > 0:
            logger.info(f"[DNSMOS] Using CUDA Provider with device_id: {device}")
            providers = [("CUDAExecutionProvider", {"device_id": device})]
        else:
            providers = ["CPUExecutionProvider"]

        self.p835_sess = ort.InferenceSession(
            (root_dir / "external" / "DNSMOS" / "sig_bak_ovr.onnx").as_posix(),
            providers=providers,
        )

        self.p808_sess = ort.InferenceSession(
            (root_dir / "external" / "DNSMOS" / "model_v8.onnx").as_posix(),
            providers=providers,
        )

        self.input_sr = input_sr

    def audio_melspec(self, audio, n_mels=120, frame_size=320, hop_length=160, sr=16000, to_db=True):
        mel_spec = librosa.feature.melspectrogram(
            y=audio, sr=sr, n_fft=frame_size + 1, hop_length=hop_length, n_mels=n_mels
        )
        if to_db:
            mel_spec = (librosa.power_to_db(mel_spec, ref=np.max) + 40) / 40  # type: ignore
        return mel_spec.T

    def get_polyfit_val(self, sig, bak, ovr, is_personalized_MOS=False):
        if is_personalized_MOS:
            p_ovr = np.poly1d([-0.00533021, 0.005101, 1.18058466, -0.11236046])
            p_sig = np.poly1d([-0.01019296, 0.02751166, 1.19576786, -0.24348726])
            p_bak = np.poly1d([-0.04976499, 0.44276479, -0.1644611, 0.96883132])
        else:
            p_ovr = np.poly1d([-0.06766283, 1.11546468, 0.04602535])
            p_sig = np.poly1d([-0.08397278, 1.22083953, 0.0052439])
            p_bak = np.poly1d([-0.13166888, 1.60915514, -0.39604546])

        sig_poly = p_sig(sig)
        bak_poly = p_bak(bak)
        ovr_poly = p_ovr(ovr)

        return sig_poly, bak_poly, ovr_poly

    def __call__(self, audio, return_p808=True):
        if audio.ndim != 1:
            audio = audio.reshape(-1)

        if torch.is_tensor(audio):
            audio = audio.detach().cpu().numpy()

        SAMPLERATE = 16000
        INPUT_LENGTH = 9.01

        if self.input_sr != 16000:
            audio = librosa.resample(audio, orig_sr=self.input_sr, target_sr=SAMPLERATE)

        len_samples = int(INPUT_LENGTH * SAMPLERATE)

        while len(audio) < len_samples:
            audio = np.append(audio, audio)

        num_hops = int(np.floor(len(audio) / SAMPLERATE) - INPUT_LENGTH) + 1

        hop_len_samples = SAMPLERATE
        predicted_mos_sig_seg = []
        predicted_mos_bak_seg = []
        predicted_mos_ovr_seg = []

        predicted_p808_mos = []

        for idx in range(num_hops):
            audio_seg = audio[int(idx * hop_len_samples) : int((idx + INPUT_LENGTH) * hop_len_samples)]
            if len(audio_seg) < len_samples:
                continue

            input_features = np.array(audio_seg).astype("float32")[np.newaxis, :]
            oi = {"input_1": input_features}

            if return_p808:
                p808_input_features = np.array(self.audio_melspec(audio=audio_seg[:-160])).astype("float32")[
                    np.newaxis, :, :
                ]
                p808_oi = {"input_1": p808_input_features}
                p808_mos = self.p808_sess.run(None, p808_oi)[0][0][0]
                predicted_p808_mos.append(p808_mos)

            mos_sig_raw, mos_bak_raw, mos_ovr_raw = self.p835_sess.run(None, oi)[0][0]
            mos_sig, mos_bak, mos_ovr = self.get_polyfit_val(mos_sig_raw, mos_bak_raw, mos_ovr_raw)

            predicted_mos_ovr_seg.append(mos_ovr)
            predicted_mos_sig_seg.append(mos_sig)
            predicted_mos_bak_seg.append(mos_bak)

        out = [
            Metric(name="OVRL", value=np.mean(predicted_mos_ovr_seg)),
            Metric(name="SIG", value=np.mean(predicted_mos_sig_seg)),
            Metric(name="BAK", value=np.mean(predicted_mos_bak_seg)),
        ]

        if return_p808:
            out.append(Metric(name="p808", value=np.mean(predicted_p808_mos)))



from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, NamedTuple, Union

import safetensors.torch
import torch
import torch.nn as nn
from simple_parsing.helpers import Serializable


@dataclass
class LoraArgs(Serializable):
    rank: int
    scaling: float

    def __post_init__(self):
        assert self.rank > 0
        assert self.scaling > 0.0


class LoRALinear(nn.Module):
    def __init__(self, in_features: int, out_features: int, rank: int, scaling: float, bias: bool = False):
        self.in_features = in_features
        self.out_features = out_features
        assert not bias
        self.bias = bias
        self.rank = rank
        self.scaling = scaling

        self.lora_A = nn.Linear(self.in_features, self.rank, bias=self.bias)
        self.lora_B = nn.Linear(self.rank, self.out_features, bias=self.bias)

        self.linear = nn.Linear(self.in_features, self.out_features, bias=self.bias)

        # make sure no LoRA weights are marked as "missing" in load_state_dict
        def ignore_missing_keys(m: nn.Module, incompatible_keys: NamedTuple):
            incompatible_keys.missing_keys[:] = []

        self.register_load_state_dict_post_hook(ignore_missing_keys)

    def forward(self, x: torch.Tensor):
        lora = self.lora_B(self.lora_A(x))
        return self.linear(x) + self.scaling * lora

    def _load_from_state_dict(self, state_dict: Dict[str, Any], prefix: str, *args: Any, **kwargs: Any):
        """Reimplement the load_from_state_dict method to handle the LoRA weights.

        `nn.Module.load_state_dict()` calls this method to load the state_dict of the current module.
        We can reimplement this method to map the keys of the state_dict to the correct weights of the module.

        Args:
            state_dict: _description_
            prefix: a string that is used to prefix the keys of the state_dict. Default is an empty string.
        """
        key_name = prefix + "weight"

        # full checkpoint
        # if key_name in state_dict, it means that this is we first time we load this LoRALinear
        # if key_name is not in state_dict, it means that we are resuming from a checkpoint during training LoRALinear
        if key_name in state_dict:
            w_ref = state_dict[key_name]

            # load frozen weights
            state_dict = {
                "linear.weight": w_ref,
                "lora_A.weight": torch.zeros_like(self.lora_A.weight, device=w_ref.device, dtype=w_ref.dtype),
                "lora_B.weight": torch.zeros_like(self.lora_B.weight, device=w_ref.device, dtype=w_ref.dtype),
            }

            self.load_state_dict(state_dict, assign=True, strict=True)


class LoRALoaderMixin:
    def load_lora(self, lora_path: Union[Path, str], scaling: float = 2.0):
        state_dict = safetensors.torch.load_file(lora_path)

        self._load_lora_state_dict(state_dict, scaling=scaling)

    def _load_lora_state_dict(self, lora_state_dict: Dict[str, torch.Tensor], scaling: float = 2.0):
        lora_dtypes = {p.dtype for p in lora_state_dict.values()}
        pass



import torch
import torch.nn as nn
import torch.nn.functional as F
from einops import rearrange

from audiozen.models.base_model import BaseModel


class ResidualLSTM(BaseModel):
    def __init__(self, input_size, output_size, hidden_size, use_activation=True):
        super().__init__()
        self.in_conv = nn.Conv1d(input_size, hidden_size, 1, bias=False)
        self.lstm_1 = nn.LSTM(hidden_size, hidden_size, batch_first=True)
        self.lstm_2 = nn.LSTM(hidden_size, hidden_size, batch_first=True)
        self.linear = nn.Linear(hidden_size, output_size)

        self.use_activation = use_activation
        self.norm_1 = nn.LayerNorm(hidden_size)
        self.norm_2 = nn.LayerNorm(output_size)

    def forward(self, input):
        output = self.in_conv(input)

        output = rearrange(output, "b f t -> b t f")
        output = output + self.norm_1(self.lstm_1(output)[0])
        output = output + self.norm_1(self.lstm_2(output)[0])
        output = self.norm_2(self.linear(output))

        if self.use_activation:
            output = F.relu(output)

        output = rearrange(output, "b t f -> b f t")

        return output


if __name__ == "__main__":
    model = ResidualLSTM(257, 257, 512)
    input = torch.rand(1, 257, 100)
    output = model(input)
    print(output.shape)



import torch
import torch.nn as nn


class SharedGroupGRULayer(nn.Module):
    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        is_first_layer,
        num_groups: int,
        batch_first: bool = True,
        bias: bool = True,
        dropout: float = 0,
        bidirectional: bool = False,
    ):
        super().__init__()
        assert hidden_size % num_groups == 0, "Should be divided by input size"

        if not is_first_layer:
            assert input_size % num_groups == 0

        kwargs = {
            "bias": bias,
            "batch_first": batch_first,
            "dropout": dropout,
            "bidirectional": bidirectional,
        }

        self.is_first_layer = is_first_layer

        if self.is_first_layer:
            self.input_size = input_size
        else:
            self.input_size = input_size // num_groups

        self.hidden_size = hidden_size // num_groups
        self.out_size = hidden_size

        self.bidirectional = bidirectional
        self.num_directions = 2 if bidirectional else 1

        self.num_groups = num_groups
        self.batch_first = batch_first

        # Several small GRUs
        self.GRU_layers = nn.ModuleList(
            (nn.GRU(self.input_size, self.hidden_size, **kwargs) for _ in range(num_groups))
        )

    def initialize_h0(self, batch_size: int = 1, device: torch.device = torch.device("cpu")):
        return torch.zeros(
            self.num_groups * self.num_directions,
            batch_size,
            self.hidden_size,
            device=device,
        )

    def forward(self, input, h0=None):
        """

        Args:
            input: [B, T, F] if batch_first else [T, B, I], B: batch_size, F: input_size
            h0: [G*D, B, H], where G: groups, D: num_directions, H: hidden_size

        Returns:

        """
        if h0 is None:
            dim0, dim1 = input.shape[:2]  # [B, T]
            batch_size = dim0 if self.batch_first else dim1
            h0 = self.initialize_h0(batch_size, device=input.device)

        outputs = []
        output_states = []

        # 实际上并没有并行的哈，循环执行四次，将大型矩阵运算变成了四个小矩阵矩阵运算后的加法
        for i, GRU_layer in enumerate(self.GRU_layers):
            o, s = GRU_layer(
                input if self.is_first_layer else input[..., i * self.input_size : (i + 1) * self.input_size],
                h0[i * self.num_directions : (i + 1) * self.num_directions].detach(),
            )
            outputs.append(o)
            output_states.append(s)

        output = torch.cat(outputs, dim=-1)
        hidden_state = torch.cat(output_states, dim=0)

        return output, hidden_state


class SharedGroupGRU(nn.Module):
    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        num_layers: int = 1,
        num_groups: int = 4,
        bias: bool = True,
        batch_first: bool = True,
        dropout: float = 0,
        bidirectional: bool = False,
        shuffle: bool = True,
        add_outputs: bool = False,
    ):
        super().__init__()

        kwargs = {
            "num_groups": num_groups,
            "bias": bias,
            "batch_first": batch_first,
            "dropout": dropout,
            "bidirectional": bidirectional,
        }

        assert hidden_size % num_groups == 0, "Should be divided by input size"

        self.num_groups = num_groups
        self.num_layers = num_layers
        self.batch_first = batch_first
        self.bidirectional = bidirectional
        self.num_directions = 2 if bidirectional else 1

        self.hidden_size = hidden_size // num_groups

        if num_groups == 1:
            shuffle = False  # Fully connected, no need to shuffle

        self.shuffle = shuffle
        self.add_outputs = add_outputs

        self.GRUs = nn.ModuleList()

        self.GRUs.append(  # First layer only
            SharedGroupGRULayer(input_size, hidden_size, True, **kwargs)
        )
        for _ in range(1, num_layers):  # Other layers
            self.GRUs.append(SharedGroupGRULayer(hidden_size, hidden_size, False, **kwargs))

    def initialize_h0(self, batch_size: int, device: torch.device = torch.device("cpu")):
        return torch.zeros(
            (
                self.num_layers * self.num_groups * self.num_directions,
                batch_size,
                self.hidden_size,
            ),
            device=device,
        )

    def forward(self, input, state=None):
        """

        Args:
            input: [B, T, F]
            state: [B, F]

        Returns:

        """
        dim0, dim1, _ = input.shape
        batch_size = dim0 if self.batch_first else dim1

        if state is None:
            state = self.initialize_h0(batch_size, input.device)

        output = torch.zeros(
            dim0,
            dim1,
            self.hidden_size * self.num_directions * self.num_groups,
            device=input.device,
        )

        output_states = []
        h = self.num_groups * self.num_directions

        for i, GRU in enumerate(self.GRUs):
            input, hidden_state = GRU(input, state[i * h : (i + 1) * h])
            output_states.append(hidden_state)

            if self.shuffle and i < self.num_layers - 1:  # TODO: Not the final layer, but WHY?
                input = input.view(dim0, dim1, -1, self.num_groups).transpose(2, 3).reshape(dim0, dim1, -1)

            if self.add_outputs:
                output += input  # TODO: WHY?
            else:
                output = input

        output_state = torch.cat(output_states, dim=0)
        return output, output_state


class GroupGRULayer(nn.Module):
    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        num_groups: int,
        batch_first: bool = True,
        bias: bool = True,
        dropout: float = 0,
        bidirectional: bool = False,
    ):
        super().__init__()
        assert input_size % num_groups == 0 and hidden_size % num_groups == 0, "Should be divided by input size"

        kwargs = {
            "bias": bias,
            "batch_first": batch_first,
            "dropout": dropout,
            "bidirectional": bidirectional,
        }

        self.input_size = input_size // num_groups
        self.hidden_size = hidden_size // num_groups
        self.out_size = hidden_size

        self.bidirectional = bidirectional
        self.num_directions = 2 if bidirectional else 1

        self.num_groups = num_groups
        self.batch_first = batch_first

        # Several small GRUs
        self.GRU_layers = nn.ModuleList(
            (nn.GRU(self.input_size, self.hidden_size, **kwargs) for _ in range(num_groups))
        )

    def initialize_h0(self, batch_size: int = 1, device: torch.device = torch.device("cpu")):
        return torch.zeros(
            self.num_groups * self.num_directions,
            batch_size,
            self.hidden_size,
            device=device,
        )

    def forward(self, input, h0=None):
        """

        Args:
            input: [B, T, F] if batch_first else [T, B, I], B: batch_size, F: input_size
            h0: [G*D, B, H], where G: groups, D: num_directions, H: hidden_size

        Returns:

        """
        if h0 is None:
            dim0, dim1 = input.shape[:2]  # [B, T]
            batch_size = dim0 if self.batch_first else dim1
            h0 = self.initialize_h0(batch_size, device=input.device)

        outputs = []
        output_states = []

        # 实际上并没有并行的哈，循环执行四次，将大型矩阵运算变成了四个小矩阵矩阵运算后的加法
        for i, GRU_layer in enumerate(self.GRU_layers):
            o, s = GRU_layer(
                input[..., i * self.input_size : (i + 1) * self.input_size],  # iterative sub_input
                h0[i * self.num_directions : (i + 1) * self.num_directions].detach(),
            )
            outputs.append(o)
            output_states.append(s)

        output = torch.cat(outputs, dim=-1)
        h = torch.cat(output_states, dim=0)

        return output, h


class GroupGRU(nn.Module):
    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        num_layers: int = 1,
        num_groups: int = 4,
        bias: bool = True,
        batch_first: bool = True,
        dropout: float = 0,
        bidirectional: bool = False,
        shuffle: bool = True,
        add_outputs: bool = False,
    ):
        super().__init__()

        kwargs = {
            "num_groups": num_groups,
            "bias": bias,
            "batch_first": batch_first,
            "dropout": dropout,
            "bidirectional": bidirectional,
        }

        assert input_size % num_groups == 0 and hidden_size % num_groups == 0, "Should be divided by input size"

        self.num_groups = num_groups
        self.num_layers = num_layers
        self.batch_first = batch_first
        self.bidirectional = bidirectional
        self.num_directions = 2 if bidirectional else 1

        self.hidden_size = hidden_size // num_groups

        if num_groups == 1:
            shuffle = False  # Fully connected, no need to shuffle

        self.shuffle = shuffle
        self.add_outputs = add_outputs

        self.GRUs = nn.ModuleList()

        self.GRUs.append(  # First layer only
            GroupGRULayer(input_size, hidden_size, **kwargs)
        )
        for _ in range(1, num_layers):  # Other layers
            self.GRUs.append(GroupGRULayer(hidden_size, hidden_size, **kwargs))

    def initialize_h0(self, batch_size: int, device: torch.device = torch.device("cpu")):
        return torch.zeros(
            (
                self.num_layers * self.num_groups * self.num_directions,
                batch_size,
                self.hidden_size,
            ),
            device=device,
        )

    def forward(self, input, state=None):
        """

        Args:
            input: [B, T, F]
            state: [B, F]

        Returns:

        """
        dim0, dim1, _ = input.shape
        batch_size = dim0 if self.batch_first else dim1

        if state is None:
            state = self.initialize_h0(batch_size, input.device)

        output = torch.zeros(
            dim0,
            dim1,
            self.hidden_size * self.num_directions * self.num_groups,
            device=input.device,
        )

        output_states = []
        h = self.num_groups * self.num_directions

        for i, GRU in enumerate(self.GRUs):
            input, hidden_state = GRU(input, state[i * h : (i + 1) * h])
            output_states.append(hidden_state)

            if self.shuffle and i < self.num_layers - 1:  # TODO: Not the final layer, but WHY?
                input = input.view(dim0, dim1, -1, self.num_groups).transpose(2, 3).reshape(dim0, dim1, -1)

            if self.add_outputs:
                output += input  # TODO: WHY?
            else:
                output = input

        output_state = torch.cat(output_states, dim=0)
        return output, output_state


class GroupedLinear(nn.Module):
    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        num_groups: int = 1,
        shuffle: bool = True,
    ):
        super().__init__()
        assert input_size % num_groups == 0 and hidden_size % num_groups == 0, "Should be divided by input size"

        self.groups = num_groups

        self.input_size = input_size // num_groups
        self.hidden_size = hidden_size // num_groups

        if num_groups == 1:
            shuffle = False

        self.shuffle = shuffle
        self.layers = nn.ModuleList(nn.Linear(self.input_size, self.hidden_size) for _ in range(num_groups))

    def forward(self, x):
        outputs = []

        for i, layer in enumerate(self.layers):
            outputs.append(layer(x[..., i * self.input_size : (i + 1) * self.input_size]))

        output = torch.cat(outputs, dim=-1)

        if self.shuffle:
            orig_shape = output.shape
            output = output.view(-1, self.hidden_size, self.groups).transpose(-1, -2).reshape(orig_shape)

        return output


if __name__ == "__main__":
    g = 10  # groups
    h = 400  # hidden_size
    i = 33  # input_size
    b = 32  # batch_size
    t = 100  # time_steps
    m = SharedGroupGRULayer(i, h, True, g, batch_first=True)
    input = torch.randn((b, t, i))
    h0 = m.initialize_h0(b)
    assert list(h0.shape) == [g, b, h // g]
    out, hout = m(input, h0)
    print(out.shape, hout.shape)

    # now grouped gru
    num = 5
    m = SharedGroupGRU(i, h, num, g, batch_first=True, shuffle=True)
    h0 = m.initialize_h0(b)
    assert list(h0.shape) == [num * g, b, h // g]
    out, hout = m(input, h0)
    print(out.shape, hout.shape)



from dataclasses import dataclass

from simple_parsing import Serializable


@dataclass
class ModelArgs(Serializable):
    sr: int = 44100
    n_fft: int = 2048
    win_length: int = 2048
    hop_length: int = 512

    enc_dim: int = 1025
    feat_dim: int = 128
    num_channels: int = 1
    num_layer: int = 1
    num_repeat: int = 6
    dropout: float = 0.0



import math
from functools import partial
from typing import Optional

from torch.optim import Optimizer
from torch.optim.lr_scheduler import LambdaLR


def _get_constant_schedule_with_warmup_lr_lambda(current_step: int, *, num_warmup_steps: int):
    if current_step < num_warmup_steps:
        return float(current_step) / float(max(1.0, num_warmup_steps))
    return 1.0


def get_constant_schedule_with_warmup(optimizer: Optimizer, num_warmup_steps: int, last_epoch: int = -1):
    lr_lambda = partial(_get_constant_schedule_with_warmup_lr_lambda, num_warmup_steps=num_warmup_steps)
    return LambdaLR(optimizer, lr_lambda, last_epoch=last_epoch)


def _get_linear_schedule_with_warmup_lr_lambda(current_step: int, *, num_warmup_steps: int, num_training_steps: int):
    if current_step < num_warmup_steps:
        return float(current_step) / float(max(1, num_warmup_steps))
    return max(0.0, float(num_training_steps - current_step) / float(max(1, num_training_steps - num_warmup_steps)))


def get_linear_schedule_with_warmup(optimizer, num_warmup_steps, num_training_steps, last_epoch=-1):
    lr_lambda = partial(
        _get_linear_schedule_with_warmup_lr_lambda,
        num_warmup_steps=num_warmup_steps,
        num_training_steps=num_training_steps,
    )
    return LambdaLR(optimizer, lr_lambda, last_epoch)


def get_cosine_with_min_lr_schedule_with_warmup(
    optimizer: Optimizer,
    num_warmup_steps: int,
    num_training_steps: int,
    num_cycles: float = 0.5,
    last_epoch: int = -1,
    initial_lr: float = 0.0,
    min_lr: Optional[float] = None,
) -> LambdaLR:
    """
    Create a learning rate scheduler that linearly increases learning rate from 0 to `lr` over `num_warmup_steps`, then
    decreases learning rate from `lr` to 0.0 on a cosine schedule over the remaining `num_training_steps - num_warmup_steps`
    steps (assuming in num_cycles=0.5).

    Args:
        optimizer: The optimizer for which to schedule the learning rate.
        num_warmup_steps: The number of steps for which to linearly increase the learning rate.
        num_training_steps: The total number of training steps.
        num_cycles: _description_. Defaults to 0.5.
        last_epoch: _description_. Defaults to -1.

    Returns:
        torch.optim.lr_scheduler.LambdaLR with the appropriate schedule.
    """
    min_lr_rate = min_lr / initial_lr

    def lr_lambda(current_step: int) -> float:
        # linear warmup phase
        if current_step < num_warmup_steps:
            return float(current_step) / float(max(1, num_warmup_steps))

        # after warmup
        progress = float(current_step - num_warmup_steps) / float(max(1, num_training_steps - num_warmup_steps))

        # move cosine curve to the top and scale it down from 0 to 1
        cosine_lr_multiple = 0.5 * (1.0 + math.cos(math.pi * float(num_cycles) * 2.0 * progress))

        # scale down to min_lr_rate
        cosine_lr_multiple = cosine_lr_multiple * (1.0 - min_lr_rate) + min_lr_rate

        return max(0.0, cosine_lr_multiple)

    return LambdaLR(optimizer, lr_lambda, last_epoch=last_epoch)



from dataclasses import dataclass
from functools import partial
from typing import Optional

import torch
import torch.nn as nn
from einops import rearrange
from simple_parsing import Serializable
from torch.nn import functional

from audiozen.acoustics.audio_feature import stft


def exists(val):
    return val is not None


def default(val, d):
    return val if exists(val) else d


@dataclass
class ModelArgs(Serializable):
    n_fft: int = 512
    hop_length: int = 256
    win_length: int = 512

    num_freqs: int = 257
    look_ahead: int = 2
    fb_num_neighbors: int = 0
    fb_model_hidden_size: int = 512
    fb_num_layers: int = 2
    fb_output_activate_function: str = "ReLU"
    sb_num_neighbors: int = 15
    sb_output_activate_function: str = "ReLU"
    sb_model_hidden_size: int = 384

    patch_merge_layer_index: Optional[int] = None
    hidden_size: int = 512
    num_freq_out = 2


class FreqMerger(nn.Module):
    def __init__(self, args: ModelArgs):
        super().__init__()

        self.scale = args.hidden_size**-0.5
        self.norm = nn.LayerNorm(args.hidden_size)
        self.queries = nn.Parameter(torch.randn(args.num_freq_out, args.hidden_size))

    def forward(self, x):
        # Normalize the input
        x = self.norm(x)  # [B, T, F_in, H]

        # Compute the similarity between the input and the queries
        sim = self.queries @ x.t()  # [F_out, H] * [B, T, H, F_in] = [B, T, F_out, F_in]
        sim *= self.scale

        # Compute the attention
        attn = torch.softmax(sim, dim=-1)  # [B, T, F_out, F_in]

        return attn @ x  # [B, T, F_out, F_in] @ [B, T, F_in, H] = [B, T, F_out, H]


class Model(nn.Module):
    def __init__(self, args: ModelArgs):
        super().__init__()
        self.fb_layer_norm = nn.LayerNorm(args.num_freqs)

        self.fb_rnn = nn.LSTM(
            input_size=args.num_freqs,
            hidden_size=args.fb_model_hidden_size,
            num_layers=2,
            batch_first=True,
            bidirectional=False,
            proj_size=args.num_freqs,
        )

        self.sb_layer_norm = nn.LayerNorm((args.sb_num_neighbors * 2 + 1) + (args.fb_num_neighbors * 2 + 1))

        self.sb_rnn = nn.LSTM(
            input_size=(args.sb_num_neighbors * 2 + 1) + (args.fb_num_neighbors * 2 + 1),
            hidden_size=args.sb_model_hidden_size,
            num_layers=2,
            batch_first=True,
            bidirectional=False,
            proj_size=2,
        )

        self.patch_merge_layer_index = (
            default(args.patch_merge_layer_index, 4 // 2) - 1
        )  # default to mid-way through transformer, as shown in paper

        self.stft = partial(stft, n_fft=args.n_fft, hop_length=args.hop_length, win_length=args.win_length)

        self.args = args

    @staticmethod
    def freq_unfold_3d(input, num_neighbors):
        """Split overlapped subband units from the frequencies axis.

        Args:
            input: three-dimension input with the shape [B, T, F]
            num_neighbors: number of neighbors in each side for each subband unit.

        Returns:
            Overlapped sub-band units specified as [B, N, T, F_s], where `F_s` represents the frequency axis of
            each sub-band unit and `N` is the number of sub-band unit, e.g., [2, 161, 1, 100, 19].
        """
        assert input.dim() == 3, f"The dim of the input is {input.dim()}. It should be three dim."

        # Convert to batched image-like tensors
        input = rearrange(input, "b t f -> b 1 f t")
        batch_size, num_channels, num_freqs, num_frames = input.size()

        # No change to the input if the number of neighbors is less than or equal to 0
        if num_neighbors <= 0:
            return rearrange(input, "b 1 f t -> b f t 1")  # where `f` is equal to the number of sub-band units

        # Pad the top and bottom of the original spectrogram
        output = functional.pad(input, [0, 0, num_neighbors, num_neighbors], mode="reflect")  # [B * C, 1, F, T]

        # Unfold the spectrogram into sub-band units by sliding along the frequency axis
        # [B * C, 1, F, T] => [B * C, sub_band_unit_size, num_frames, N], N is equal to the number of frequencies.
        sub_band_unit_size = num_neighbors * 2 + 1
        output = functional.unfold(output, kernel_size=(sub_band_unit_size, num_frames))
        assert output.shape[-1] == num_freqs, f"n_freqs != N (sub_band), {num_freqs} != {output.shape[-1]}"

        # Split the dimension of the unfolded feature
        output = rearrange(output, "b (fs t) n -> b n t fs", fs=sub_band_unit_size)

        return output

    def forward(self, input_features):
        """
        Args:
            noisy_mag: noisy magnitude spectrogram

        Returns:
            The real part and imag part of the enhanced spectrogram

        Shapes:
            noisy_mag: [B, 1, F, T]
            return: [B, 2, F, T]
        """
        assert input_features.dim() == 4, f"The dim of the input is {input_features.dim()}. It should be four dim."
        input_features = functional.pad(input_features, [0, self.args.look_ahead])  # Pad the look ahead
        batch_size, num_channels, num_freqs, num_frames = input_features.size()
        assert num_channels == 1, f"{self.__class__.__name__} takes the mag feature as inputs."

        # Convert to channel last format
        input_features = rearrange(input_features, "b c f t -> b t (c f)")

        # Fullband model
        fb_input = self.fb_layer_norm(input_features)
        fb_output, _ = self.fb_rnn(fb_input)  # [B, T, F]

        # Unfold fullband model's output, [B, N=F, T, F_s]. N is the number of sub-band units
        fb_output_unfolded = self.freq_unfold_3d(fb_output, num_neighbors=self.args.fb_num_neighbors)
        # Unfold noisy spectrogram, [B, N=F, T, F_s]
        noisy_mag_unfolded = self.freq_unfold_3d(input_features, num_neighbors=self.args.sb_num_neighbors)

        # Concatenation, [B, F, T, (F_s + F_f)]
        sb_input = torch.cat([noisy_mag_unfolded, fb_output_unfolded], dim=-1)
        sb_input = self.sb_layer_norm(sb_input)
        sb_input = rearrange(sb_input, "b f t fs -> (b f) t fs")
        sb_mask, _ = self.sb_rnn(sb_input)  #  [(B * F), T, 2]
        sb_mask = rearrange(sb_mask, "(b f) t ri -> b ri f t", b=batch_size, f=num_freqs)

        output = sb_mask[:, :, :, self.args.look_ahead :]
        return output


if __name__ == "__main__":
    with torch.no_grad():
        noisy_mag = torch.rand(1, 1, 257, 63)
        model = Model(ModelArgs())
        print(model(noisy_mag).shape)



from dataclasses import dataclass
from typing import Optional

from simple_parsing.helpers import Serializable

from audiozen.models.transformer.lora import LoraArgs


@dataclass
class ModelArgs(Serializable):
    dim: int
    n_layers: int
    n_heads: int
    head_dim: int
    hidden_dim: int
    norm_eps: float

    # For rotary embeddings. If not set, will be inferred
    rope_theta: Optional[float] = None

    lora: Optional[LoraArgs] = None  # If this is set, we will load LoRA linear layers instead of linear layers.



import torch
import torch.nn as nn
import torch.nn.init as init
from torch.nn import functional

from audiozen.constant import EPSILON


class BaseModel(nn.Module):
    def __init__(self):
        super(BaseModel, self).__init__()

    @staticmethod
    def freq_unfold(input, num_neighbors):
        """Split the overlapped subband units along the frequency axis.

        Args:
            input: four-dimension input with the shape [B, C, F, T].
            num_neighbors: number of neighbors in each side for each subband unit.

        Returns:
            Overlapped sub-band units specified as [B, N, C, F_s, T], where `F_s` represents the frequency axis of
            each sub-band unit and `N` is the number of sub-band unit, e.g., [2, 161, 1, 19, 100].
        """
        assert input.dim() == 4, f"The dim of the input is {input.dim()}. It should be four dim."
        batch_size, num_channels, num_freqs, num_frames = input.size()

        if num_neighbors <= 0:  # No change to the input
            return input.permute(0, 2, 1, 3).reshape(batch_size, num_freqs, num_channels, 1, num_frames)

        output = input.reshape(batch_size * num_channels, 1, num_freqs, num_frames)  # [B * C, 1, F, T]
        sub_band_unit_size = num_neighbors * 2 + 1

        # Pad the top and bottom of the original spectrogram
        output = functional.pad(output, [0, 0, num_neighbors, num_neighbors], mode="reflect")  # [B * C, 1, F, T]

        # Unfold the spectrogram into sub-band units
        # [B * C, 1, F, T] => [B * C, sub_band_unit_size, num_frames, N], N is equal to the number of frequencies.
        output = functional.unfold(output, kernel_size=(sub_band_unit_size, num_frames))  # move on the F and T axes
        assert output.shape[-1] == num_freqs, f"n_freqs != N (sub_band), {num_freqs} != {output.shape[-1]}"

        # Split the dimension of the unfolded feature
        output = output.reshape(batch_size, num_channels, sub_band_unit_size, num_frames, num_freqs)
        output = output.permute(0, 4, 1, 2, 3).contiguous()  # [B, N, C, F_s, T]

        return output

    @staticmethod
    def unfold_time(input, kernel_size, stride):
        """Unfold (like sliding window) input tensor along the time axis.

        Shapes:
            input: [B, C, F, T]. PyTorch only support 4-D (image-like) tensor.
            kernel_size: Tuple with two values
            stride: Tuple with two values
            return: [B, C, kernel_size_1, kernel_size_2, N]
        """
        batch_size, num_channels, num_freqs, num_frames = input.shape

        # [B, C * prod(kernel_size), N]
        output = torch.nn.functional.unfold(input, kernel_size=kernel_size, stride=stride)
        num_blocks = output.shape[-1]

        # [B, C, kernel_size[0], kernel_size[1], N]
        output = output.reshape(batch_size, num_channels, *kernel_size, num_blocks)

        return output

    @staticmethod
    def time_fold(input, kernel_size, stride, output_size):
        """Fold (like sliding window) input tensor along the time axis.

        Shapes:
            input: [B, C, kernel_size[0], kernel_size[1], N]
            kernel_size: Tuple with two values
            stride: Tuple with two values
            output_size: Tuple of the spatial shape, e.g., (num_freqs, num_frames)
            return: [B, C, F, T]
        """
        batch_size, num_channels, ks_1, ks_2, num_blocks = input.shape
        input = input.reshape(batch_size, -1, num_blocks)  # [B, C * prod(kernel_size), N]

        output = torch.nn.functional.fold(
            input, kernel_size=kernel_size, stride=stride, output_size=output_size
        )  # [B, C, F, T]

        return output

    @staticmethod
    def _pad_length(tensor, win_length):
        """Pad the length of the tensor to be divisible by the win_length.

        Args:
            tensor: [*, T].
            win_length: window length.

        Returns:
            Padded tensor with the shape of [*, T_pad].
        """
        seq_len = tensor.shape[-1]
        remainder = seq_len % win_length

        pad_size = win_length - remainder if remainder > 0 else 0

        if pad_size > 0:
            tensor = torch.cat(
                (
                    tensor,
                    torch.zeros([*(tensor.shape[:-1]), pad_size], device=tensor.device),
                ),
                dim=-1,
            )

        return tensor, pad_size

    @staticmethod
    def _reduce_complexity_separately(sub_band_input, full_band_output, device):
        """Group Dropout for FullSubNet.

        Args:
            sub_band_input: [60, 257, 1, 33, 200]
            full_band_output: [60, 257, 1, 3, 200]
            device:

        Notes:
            1. 255 and 256 freq not able to be trained
            2. batch size 应该被 3 整除，否则最后一部分 batch 内的频率无法很好的训练

        Returns:
            [60, 85, 1, 36, 200]
        """
        batch_size = full_band_output.shape[0]
        n_freqs = full_band_output.shape[1]
        sub_batch_size = batch_size // 3
        final_selected = []

        for idx in range(3):
            # [0, 60) => [0, 20)
            sub_batch_indices = torch.arange(idx * sub_batch_size, (idx + 1) * sub_batch_size, device=device)
            full_band_output_sub_batch = torch.index_select(full_band_output, dim=0, index=sub_batch_indices)
            sub_band_output_sub_batch = torch.index_select(sub_band_input, dim=0, index=sub_batch_indices)

            # Avoid to use padded value (first freq and last freq)
            # i = 0, (1, 256, 3) = [1, 4, ..., 253]
            # i = 1, (2, 256, 3) = [2, 5, ..., 254]
            # i = 2, (3, 256, 3) = [3, 6, ..., 255]
            freq_indices = torch.arange(idx + 1, n_freqs - 1, step=3, device=device)
            full_band_output_sub_batch = torch.index_select(full_band_output_sub_batch, dim=1, index=freq_indices)
            sub_band_output_sub_batch = torch.index_select(sub_band_output_sub_batch, dim=1, index=freq_indices)

            # ([30, 85, 1, 33 200], [30, 85, 1, 3, 200]) => [30, 85, 1, 36, 200]

            final_selected.append(torch.cat([sub_band_output_sub_batch, full_band_output_sub_batch], dim=-2))

        return torch.cat(final_selected, dim=0)

    @staticmethod
    def forgetting_norm(input, sample_length=192):
        """Using the mean value of the near frames to normalization.

        Args:
            input: feature
            sample_length: length of the training sample, used for calculating smooth factor

        Returns:
            normed feature

        Shapes:
            input: [B, C, F, T]
            sample_length_in_training: 192
        """
        assert input.ndim == 4
        batch_size, num_channels, num_freqs, num_frames = input.size()
        input = input.reshape(batch_size, num_channels * num_freqs, num_frames)

        eps = EPSILON
        mu = 0
        alpha = (sample_length - 1) / (sample_length + 1)

        mu_list = []
        # TODO bugs
        # TODO 是否需要修改为包含F的计算方法？
        for frame_idx in range(num_frames):
            if frame_idx < sample_length:
                alp = torch.min(torch.tensor([(frame_idx - 1) / (frame_idx + 1), alpha]))
                mu = alp * mu + (1 - alp) * torch.mean(input[:, :, frame_idx], dim=1).reshape(batch_size, 1)  # [B, 1]
            else:
                current_frame_mu = torch.mean(input[:, :, frame_idx], dim=1).reshape(batch_size, 1)  # [B, 1]
                mu = alpha * mu + (1 - alpha) * current_frame_mu

            mu_list.append(mu)

            # print("input", input[:, :, idx].min(), input[:, :, idx].max(), input[:, :, idx].mean())
            # print(f"alp {idx}: ", alp)
            # print(f"mu {idx}: {mu[128, 0]}")

        mu = torch.stack(mu_list, dim=-1)  # [B, 1, T]
        output = input / (mu + eps)

        output = output.reshape(batch_size, num_channels, num_freqs, num_frames)
        return output

    @staticmethod
    def hybrid_norm(input, sample_length_in_training=192):
        """
        Args:
            input: [B, F, T]
            sample_length_in_training:

        Returns:
            [B, F, T]
        """
        assert input.ndim == 3
        device = input.device
        data_type = input.dtype
        batch_size, n_freqs, n_frames = input.size()
        eps = EPSILON

        mu = 0
        alpha = (sample_length_in_training - 1) / (sample_length_in_training + 1)
        mu_list = []
        for idx in range(input.shape[-1]):
            if idx < sample_length_in_training:
                alp = torch.min(torch.tensor([(idx - 1) / (idx + 1), alpha]))
                mu = alp * mu + (1 - alp) * torch.mean(input[:, :, idx], dim=1).reshape(batch_size, 1)  # [B, 1]
                mu_list.append(mu)
            else:
                break
        initial_mu = torch.stack(mu_list, dim=-1)  # [B, 1, T]

        step_sum = torch.sum(input, dim=1)  # [B, T]
        cumulative_sum = torch.cumsum(step_sum, dim=-1)  # [B, T]

        entry_count = torch.arange(n_freqs, n_freqs * n_frames + 1, n_freqs, dtype=data_type, device=device)
        entry_count = entry_count.reshape(1, n_frames)  # [1, T]
        entry_count = entry_count.expand_as(cumulative_sum)  # [1, T] => [B, T]

        cum_mean = cumulative_sum / entry_count  # B, T

        cum_mean = cum_mean.reshape(batch_size, 1, n_frames)  # [B, 1, T]

        # print(initial_mu[0, 0, :50])
        # print("-"*60)
        # print(cum_mean[0, 0, :50])
        cum_mean[:, :, :sample_length_in_training] = initial_mu

        return input / (cum_mean + eps)

    @staticmethod
    def offline_laplace_norm(input, return_mu=False):
        """Normalize the input with the utterance-level mean.

        Args:
            input: [B, C, F, T]

        Returns:
            [B, C, F, T]

        Notes:
            As mentioned in the paper, the offline normalization is used.
            Based on a bunch of experiments, the offline normalization have the same performance as the cumulative one and have a faster convergence than the cumulative one.
            Therefore, we use the offline normalization as the default normalization method.
        """
        # utterance-level mu
        mu = torch.mean(input, dim=list(range(1, input.dim())), keepdim=True)

        normed = input / (mu + EPSILON)

        if return_mu:
            return normed, mu
        else:
            return normed

    @staticmethod
    def cumulative_laplace_norm(input):
        """Normalize the input with the cumulative mean.

        Args:
            input: [B, C, F, T]

        Returns:
        """
        batch_size, num_channels, num_freqs, num_frames = input.size()
        input = input.reshape(batch_size * num_channels, num_freqs, num_frames)

        step_sum = torch.sum(input, dim=1)  # [B * C, F, T] => [B, T]
        cumulative_sum = torch.cumsum(step_sum, dim=-1)  # [B, T]

        entry_count = torch.arange(
            num_freqs,
            num_freqs * num_frames + 1,
            num_freqs,
            dtype=input.dtype,
            device=input.device,
        )
        entry_count = entry_count.reshape(1, num_frames)  # [1, T]
        entry_count = entry_count.expand_as(cumulative_sum)  # [1, T] => [B, T]

        cumulative_mean = cumulative_sum / entry_count  # B, T
        cumulative_mean = cumulative_mean.reshape(batch_size * num_channels, 1, num_frames)

        normed = input / (cumulative_mean + EPSILON)

        return normed.reshape(batch_size, num_channels, num_freqs, num_frames)

    @staticmethod
    def offline_gaussian_norm(input):
        """
        Zero-Norm
        Args:
            input: [B, C, F, T]

        Returns:
            [B, C, F, T]
        """
        mu = torch.mean(input, dim=list(range(1, input.dim())), keepdim=True)
        std = torch.std(input, dim=list(range(1, input.dim())), keepdim=True)

        normed = (input - mu) / (std + EPSILON)
        return normed

    @staticmethod
    def cumulative_layer_norm(input):
        """Online zero-norm.

        Args:
            input: [B, C, F, T]

        Returns:
            [B, C, F, T]
        """
        batch_size, num_channels, num_freqs, num_frames = input.size()
        input = input.reshape(batch_size * num_channels, num_freqs, num_frames)

        step_sum = torch.sum(input, dim=1)  # [B * C, F, T] => [B, T]
        step_pow_sum = torch.sum(torch.square(input), dim=1)

        cumulative_sum = torch.cumsum(step_sum, dim=-1)  # [B, T]
        cumulative_pow_sum = torch.cumsum(step_pow_sum, dim=-1)  # [B, T]

        entry_count = torch.arange(
            num_freqs,
            num_freqs * num_frames + 1,
            num_freqs,
            dtype=input.dtype,
            device=input.device,
        )
        entry_count = entry_count.reshape(1, num_frames)  # [1, T]
        entry_count = entry_count.expand_as(cumulative_sum)  # [1, T] => [B, T]

        cumulative_mean = cumulative_sum / entry_count  # [B, T]
        cumulative_var = (
            cumulative_pow_sum - 2 * cumulative_mean * cumulative_sum
        ) / entry_count + cumulative_mean.pow(2)  # [B, T]
        cumulative_std = torch.sqrt(cumulative_var + EPSILON)  # [B, T]

        cumulative_mean = cumulative_mean.reshape(batch_size * num_channels, 1, num_frames)
        cumulative_std = cumulative_std.reshape(batch_size * num_channels, 1, num_frames)

        normed = (input - cumulative_mean) / cumulative_std

        return normed.reshape(batch_size, num_channels, num_freqs, num_frames)

    def norm_wrapper(self, norm_type: str):
        if norm_type == "offline_laplace_norm":
            norm = self.offline_laplace_norm
        elif norm_type == "cumulative_laplace_norm":
            norm = self.cumulative_laplace_norm
        elif norm_type == "offline_gaussian_norm":
            norm = self.offline_gaussian_norm
        elif norm_type == "cumulative_layer_norm":
            norm = self.cumulative_layer_norm
        elif norm_type == "forgetting_norm":
            norm = self.forgetting_norm
        else:
            raise NotImplementedError(
                "You must set up a type of Norm. "
                "e.g. offline_laplace_norm, cumulative_laplace_norm, forgetting_norm, etc."
            )
        return norm

    def weight_init(self, m):
        """
        Usage:
            model = Model()
            model.apply(weight_init)
        """
        if isinstance(m, nn.Conv1d):
            init.normal_(m.weight.data)
            if m.bias is not None:
                init.normal_(m.bias.data)
        elif isinstance(m, nn.Conv2d):
            init.xavier_normal_(m.weight.data)
            if m.bias is not None:
                init.normal_(m.bias.data)
        elif isinstance(m, nn.Conv3d):
            init.xavier_normal_(m.weight.data)
            if m.bias is not None:
                init.normal_(m.bias.data)
        elif isinstance(m, nn.ConvTranspose1d):
            init.normal_(m.weight.data)
            if m.bias is not None:
                init.normal_(m.bias.data)
        elif isinstance(m, nn.ConvTranspose2d):
            init.xavier_normal_(m.weight.data)
            if m.bias is not None:
                init.normal_(m.bias.data)
        elif isinstance(m, nn.ConvTranspose3d):
            init.xavier_normal_(m.weight.data)
            if m.bias is not None:
                init.normal_(m.bias.data)
        elif isinstance(m, nn.BatchNorm1d):
            init.normal_(m.weight.data, mean=1, std=0.02)
            init.constant_(m.bias.data, 0)
        elif isinstance(m, nn.BatchNorm2d):
            init.normal_(m.weight.data, mean=1, std=0.02)
            init.constant_(m.bias.data, 0)
        elif isinstance(m, nn.BatchNorm3d):
            init.normal_(m.weight.data, mean=1, std=0.02)
            init.constant_(m.bias.data, 0)
        elif isinstance(m, nn.Linear):
            init.xavier_normal_(m.weight.data)
            init.normal_(m.bias.data)
        elif isinstance(m, nn.LSTM):
            for param in m.parameters():
                if len(param.shape) >= 2:
                    init.orthogonal_(param.data)
                else:
                    init.normal_(param.data)
        elif isinstance(m, nn.LSTMCell):
            for param in m.parameters():
                if len(param.shape) >= 2:
                    init.orthogonal_(param.data)
                else:
                    init.normal_(param.data)
        elif isinstance(m, nn.GRU):
            for param in m.parameters():
                if len(param.shape) >= 2:
                    init.orthogonal_(param.data)
                else:
                    init.normal_(param.data)
        elif isinstance(m, nn.GRUCell):
            for param in m.parameters():
                if len(param.shape) >= 2:
                    init.orthogonal_(param.data)
                else:
                    init.normal_(param.data)



from torch.optim.lr_scheduler import _LRScheduler


class CustomLRScheduler(_LRScheduler):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def get_lr(self):
        pass



import math
from collections import namedtuple

import torch
import torch.nn as nn
from torch.nn import Parameter


MemoryState = namedtuple("MemoryState", ["hx", "cx"])


def efficient_spiking_neuron(
    input_size,
    hidden_size,
    num_layers,
    shared_weights=False,
    bn=False,
    batch_first=False,
):
    """Instantiate efficient spiking networks where each spiking neuron uses the gating mechanism to control the decay
    of membrane potential.

    :param input_size:
    :param hidden_size:
    :param num_layers:
    :param shared_weights: whether weights of the gate are shared with the ones of the cell or not.
    :param bn: whether batchnorm is used or not.
    :param batch_first: Not used.
    :return:
    """
    #     # The following are not implemented.
    assert not batch_first
    # assert shared_weights
    # assert bn

    return StackedGSU(
        num_layers,
        GSULayer,
        first_layer_args=[GSUCell, input_size, hidden_size, shared_weights, bn],
        other_layer_args=[GSUCell, hidden_size, hidden_size, shared_weights, bn],
    )


class StackedGSU(nn.Module):
    # __constants__ = ["layers"]  # Necessary for iterating through self.layers

    def __init__(self, num_layers, layer, first_layer_args, other_layer_args):
        super(StackedGSU, self).__init__()
        self.layers = init_stacked_gsu(num_layers, layer, first_layer_args, other_layer_args)

    def forward(self, input, states):
        output_states = []
        output = input
        # XXX: enumerate https://github.com/pytorch/pytorch/issues/14471
        i = 0
        all_layer_output = [input]
        for rnn_layer in self.layers:
            state = states[i]
            output, out_state = rnn_layer(output, state)
            output_states += [out_state]
            all_layer_output += [output]
            i += 1
        return output, output_states, all_layer_output


def init_stacked_gsu(num_layers, layer, first_layer_args, other_layer_args):
    layers = [layer(*first_layer_args)] + [layer(*other_layer_args) for _ in range(num_layers - 1)]
    return nn.ModuleList(layers)


class GSULayer(nn.Module):
    def __init__(self, cell, *cell_args):
        super(GSULayer, self).__init__()
        self.cell = cell(*cell_args)

    def forward(self, input, state):
        inputs = input.unbind(0)
        outputs = []
        for i in range(len(inputs)):
            out, state = self.cell(inputs[i], state)
            outputs += [out]
        return torch.stack(outputs), state


class Triangle(torch.autograd.Function):
    """Spike firing activation function."""

    @staticmethod
    def forward(ctx, input, gamma=1.0):
        out = input.ge(0.0).float()
        L = torch.tensor([gamma])
        ctx.save_for_backward(input, out, L)
        return out

    @staticmethod
    def backward(ctx, grad_output):
        (input, out, others) = ctx.saved_tensors
        gamma = others[0].item()
        grad_input = grad_output.clone()
        tmp = (1 / gamma) * (1 / gamma) * ((gamma - input.abs()).clamp(min=0))
        grad_input = grad_input * tmp
        return grad_input, None


class GSUCell(nn.Module):
    def __init__(self, input_size, hidden_size, shared_weights=False, bn=False):
        super(GSUCell, self).__init__()
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.shared_weights = shared_weights
        self.use_bn = bn
        if self.shared_weights:
            self.weight_ih = Parameter(torch.empty(hidden_size, input_size))
            self.weight_hh = Parameter(torch.empty(hidden_size, hidden_size))
        else:
            self.weight_ih = Parameter(torch.empty(2 * hidden_size, input_size))
            self.weight_hh = Parameter(torch.empty(2 * hidden_size, hidden_size))
        self.bias_ih = Parameter(torch.zeros(2 * hidden_size))
        # self.bias_hh = Parameter(torch.zeros(2 * hidden_size))
        self.reset_parameters()

        # self.scale_factor = Parameter(torch.ones(hidden_size))
        if self.use_bn:
            self.batchnorm = nn.BatchNorm1d(hidden_size)

        # self.scale = torch.tensor((1 - (-1)) / (127 - (-128)))

    def reset_parameters(self):
        stdv = 1.0 / math.sqrt(self.hidden_size) if self.hidden_size > 0 else 0
        for weight in self.parameters():
            torch.nn.init.uniform_(weight, -stdv, stdv)

    def forward(self, input, state):
        hx, cx = state
        if self.shared_weights:
            weight_ih = self.weight_ih.repeat((2, 1))
            weight_hh = self.weight_hh.repeat((2, 1))
        else:
            weight_ih = self.weight_ih
            weight_hh = self.weight_hh
        gates = (
            torch.mm(input, weight_ih.t()) + self.bias_ih + torch.mm(hx, weight_hh.t())
            # + self.bias_hh
        )
        forgetgate, cellgate = gates.chunk(2, 1)
        forgetgate = torch.sigmoid(forgetgate)
        cy = forgetgate * cx + (1 - forgetgate) * cellgate
        if self.use_bn:
            cy = self.batchnorm(cy)
        hy = Triangle.apply(cy)  # replace the Tanh activation function with step function to ensure binary outputs.

        return hy, (hy, cy)


if __name__ == "__main__":
    input_size = 256
    hidden_size = 320
    num_layers = 2
    shared_weights = True
    bn = True
    batch_size = 128
    T = 100
    x = torch.rand((batch_size, input_size, T))  # [B, F, T]
    sequence_model = efficient_spiking_neuron(
        input_size=input_size,
        hidden_size=hidden_size,
        num_layers=num_layers,
        shared_weights=shared_weights,
        bn=bn,
    )

    states = [
        MemoryState(
            torch.zeros(batch_size, hidden_size, device=x.device),
            torch.zeros(batch_size, hidden_size, device=x.device),
        )
        for _ in range(num_layers)
    ]
    x = x.permute(2, 0, 1).contiguous()  # [B, F, T] => [T, B, F]
    o, _ = sequence_model(x, states)  # [T, B, F] => [T, B, F]



from dataclasses import dataclass
from functools import partial
from typing import List, Literal, Optional, Tuple, Union

import numpy as np
import torch
import torch.nn as nn
from einops import rearrange
from torch import Tensor


@dataclass
class ModelArgs:
    sr: int = 44100
    n_fft: int = 2048
    win_length: int = 2048
    hop_length: int = 512

    enc_dim: int = 1025
    feat_dim: int = 128
    num_channels: int = 1
    num_layer: int = 1
    num_repeat: int = 6
    dropout: float = 0.0


def mag_phase(complex_valued_tensor: Tensor) -> tuple[Tensor, Tensor]:
    """Get magnitude and phase of a complex-valued tensor.

    Args:
        complex_valued_tensor: complex-valued tensor.

    Returns:
        magnitude and phase spectrogram.
    """
    mag, phase = torch.abs(complex_valued_tensor), torch.angle(complex_valued_tensor)
    return mag, phase


def real_imag(complex_valued_tensor: Tensor) -> tuple[Tensor, Tensor]:
    """Get real and imag part of a complex-valued tensor.

    Args:
        complex_valued_tensor: complex-valued tensor.

    Returns:
        real and imag spectrogram.
    """
    return complex_valued_tensor.real, complex_valued_tensor.imag


def istft(
    feature: Union[Tensor, Tuple[Tensor, Tensor], List[Tensor]],
    n_fft: int,
    hop_length: int,
    win_length: int,
    window: Optional[str] = "hann",
    length: Optional[int] = None,
    input_type: Literal["mag_phase", "real_imag", "complex", "complex_view_as_real"] = "mag_phase",
) -> Tensor:
    if input_type == "real_imag":
        if isinstance(feature, (tuple, list)) and len(feature) == 2:
            real, imag = feature
            complex_valued_features = torch.complex(real=real, imag=imag)
        else:
            raise ValueError("For 'real_imag', provide a tuple or list with two elements: (real, imag).")

    elif input_type == "complex":
        if isinstance(feature, Tensor) and torch.is_complex(feature):
            complex_valued_features = feature
        else:
            raise ValueError("For 'complex', provide a complex-valued tensor.")

    elif input_type == "complex_view_as_real":
        assert isinstance(feature, Tensor), "For 'complex_view_as_real', provide a complex-valued tensor."
        if feature.ndim == 4 and feature.shape[1] == 2:
            real, imag = feature[:, 0, ...], feature[:, 1, ...]
            complex_valued_features = torch.complex(real=real, imag=imag)
        else:
            raise ValueError("For 'complex_view_as_real', provide a 4D tensor with shape [B, 2, F, T].")

    elif input_type == "mag_phase":
        if isinstance(feature, (tuple, list)) and len(feature) == 2:
            mag, phase = feature
            complex_valued_features = torch.polar(mag, phase)
        else:
            raise ValueError("For 'mag_phase', provide a tuple or list with two elements: (magnitude, phase).")

    else:
        raise ValueError(f"Unsupported input_type: {input_type}")

    if window is not None:
        window_func = getattr(torch, f"{window}_window", None)
        if window_func is None:
            raise ValueError(f"Unsupported window type: {window}")
        window_initialized = window_func(win_length, device=complex_valued_features.device)
    else:
        window_initialized = torch.ones(win_length, device=complex_valued_features.device)

    return torch.istft(
        input=complex_valued_features,
        n_fft=n_fft,
        hop_length=hop_length,
        win_length=win_length,
        window=window_initialized,
        length=length,
    )


def stft(
    y,
    n_fft,
    hop_length,
    win_length,
    window="hann",
    output_type: Literal["mag_phase", "real_imag", "complex"] | None = None,
    **kwargs,
) -> tuple[Tensor, ...] | Tensor:
    """Wrapper of the official ``torch.stft`` for single-channel and multichannel signals.

    Args:
        y (`torch.Tensor` of shape `(batch_size, num_channels, num_samples) or `(batch_size, num_samples)`):
            single-/multichannel signals.
        n_fft: num of FFT.
        hop_length: hop length.
        win_length: hanning window size.
        output_type: "mag_phase", "real_imag", "complex", or None. Defaults to None.
        kwargs: other arguments for ``torch.stft``.

    Returns:
        If the input is single-channel, return the spectrogram with shape of [B, F, T], otherwise [B, C, F, T].
        If output_type is "mag_phase", return a list of magnitude and phase spectrogram.
        If output_type is "real_imag", return a list of real and imag spectrogram.
        If output_type is None, return a list of magnitude, phase, real, and imag spectrogram.
    """
    # logger.warning("This function is deprecated. Please use `stft_v2` instead.")

    # Support [B, T] and [B, C, T]
    if y.ndim not in [2, 3]:
        raise ValueError(f"Only support single-/multi-channel signals. Received {y.ndim=}.")

    is_multi_channel = y.ndim == 3

    batch_size, *_, num_samples = y.shape

    # Compatible with multi-channel signals
    if is_multi_channel:
        y = y.reshape(-1, num_samples)

    if window is not None:
        window_func = getattr(torch, f"{window}_window")
        window = window_func(win_length, device=y.device)
    else:
        window = torch.ones(n_fft, device=y.device)

    complex_stft = torch.stft(
        input=y,
        n_fft=n_fft,
        hop_length=hop_length,
        win_length=win_length,
        window=window,
        return_complex=True,
        **kwargs,
    )

    # Reshape back to original if the input is multi-channel
    if is_multi_channel:
        complex_stft = complex_stft.reshape(batch_size, -1, *complex_stft.shape[-2:])

    if output_type == "mag_phase":
        return mag_phase(complex_stft)
    elif output_type == "real_imag":
        return real_imag(complex_stft)
    elif output_type == "complex":
        return complex_stft
    else:
        mag, phase = mag_phase(complex_stft)
        return mag, phase, complex_stft.real, complex_stft.imag


class ResRNN(nn.Module):
    def __init__(self, input_size, hidden_size, bidirectional=True, residual=True):
        super(ResRNN, self).__init__()

        self.input_size = input_size
        self.hidden_size = hidden_size
        self.residual = residual
        self.eps = 1e-6

        self.norm = nn.GroupNorm(1, input_size, self.eps)
        self.rnn = nn.LSTM(input_size, hidden_size, 1, batch_first=True, bidirectional=bidirectional)

        # linear projection layer
        self.proj = nn.Linear(hidden_size * (int(bidirectional) + 1), input_size)

    def forward(self, input):
        # input shape: batch, dim, seq

        rnn_output, _ = self.rnn(self.norm(input).transpose(1, 2).contiguous())
        rnn_output = self.proj(rnn_output.contiguous().view(-1, rnn_output.shape[2])).view(
            input.shape[0], input.shape[2], input.shape[1]
        )
        rnn_output = rnn_output.transpose(1, 2).contiguous()

        if self.residual:
            return input + rnn_output
        else:
            return rnn_output


class BSNet(nn.Module):
    def __init__(self, in_channel, num_bands=7, num_layer=1):
        super(BSNet, self).__init__()

        self.num_bands = num_bands
        self.feat_dim = in_channel // num_bands

        self.band_rnn = []
        for _ in range(num_layer):
            self.band_rnn.append(ResRNN(self.feat_dim, self.feat_dim * 2, bidirectional=True))
        self.band_rnn = nn.Sequential(*self.band_rnn)
        self.band_comm = ResRNN(self.feat_dim, self.feat_dim * 2, bidirectional=True)

    def forward(self, input):
        input, active_nband, bsnet_layer = (input[0], input[1], input[2])

        # input shape: B, nband*N, T
        B, N, T = input.shape
        # Fuse the tsvad_embedding and subband feature

        input = input.view(B * self.num_bands, self.feat_dim, -1)

        if active_nband is None:
            band_output = self.band_rnn(input).view(B, self.num_bands, -1, T)  # type: ignore
            # band comm
            band_output = band_output.permute(0, 3, 2, 1).contiguous().view(B * T, -1, self.num_bands)
            output = self.band_comm(band_output).view(B, T, -1, self.num_bands).permute(0, 3, 2, 1).contiguous()
            return output.view(B, N, T)
        else:
            band_output = self.band_rnn(input).view(B, active_nband, -1, T)  # type: ignore

            # band comm
            band_output = band_output.permute(0, 3, 2, 1).contiguous().view(B * T, -1, active_nband)
            output = self.band_comm(band_output).view(B, T, -1, active_nband).permute(0, 3, 2, 1).contiguous()

            return [output.view(B, N, T), active_nband, bsnet_layer + 1]


class Model(nn.Module):
    def __init__(self, args: ModelArgs):
        super(Model, self).__init__()
        self.args = args

        # 0-1k (100 hop), 1k-4k (250 hop), 4k-8k (500 hop), 8k-16k (1k hop), 16k-20k (2k hop), 20k-inf
        bandwidth_100 = int(np.floor(100 / (args.sr / 2.0) * args.enc_dim))
        bandwidth_250 = int(np.floor(250 / (args.sr / 2.0) * args.enc_dim))
        bandwidth_500 = int(np.floor(500 / (args.sr / 2.0) * args.enc_dim))
        bandwidth_1k = int(np.floor(1000 / (args.sr / 2.0) * args.enc_dim))
        # bandwidth_2k = int(np.floor(2000 / (args.sr / 2.0) * args.enc_dim))
        self.band_width = [bandwidth_100] * 10
        self.band_width += [bandwidth_250] * 12
        self.band_width += [bandwidth_500] * 8
        self.band_width += [bandwidth_1k] * 8
        # self.band_width += [bandwidth_2k] * 2
        self.band_width.append(int(args.enc_dim - np.sum(self.band_width)))
        self.n_band = len(self.band_width)
        print(self.band_width)

        self.eps = 1e-6
        ri_dim = 2

        self.BN = nn.ModuleList([])
        for i in range(self.n_band):
            self.BN.append(
                nn.Sequential(
                    nn.GroupNorm(1, self.band_width[i] * ri_dim * args.num_channels, self.eps),
                    nn.Conv1d(self.band_width[i] * ri_dim * args.num_channels, args.feat_dim, 1),
                )
            )

        self.separator = []
        for i in range(args.num_repeat):
            self.separator.append(BSNet(self.n_band * args.feat_dim, self.n_band, args.num_layer))
        self.separator = nn.Sequential(*self.separator)

        self.mask = nn.ModuleList([])
        for i in range(self.n_band):
            self.mask.append(
                nn.Sequential(
                    nn.GroupNorm(1, args.feat_dim, self.eps),
                    nn.Conv1d(args.feat_dim, args.feat_dim * 2, 1),
                    nn.Tanh(),
                    nn.Conv1d(args.feat_dim * 2, args.feat_dim * 2, 1),
                    nn.Tanh(),
                    nn.Conv1d(args.feat_dim * 2, self.band_width[i] * 4 * 2, 1),
                )
            )

        self.args = args
        self.stft = partial(stft, n_fft=args.n_fft, hop_length=args.hop_length, win_length=args.win_length)
        self.istft = partial(istft, n_fft=args.n_fft, hop_length=args.hop_length, win_length=args.win_length)

    def forward(self, noisy_y):
        # noisy_y: [B, C, T] or [B, T]
        # enh_y: [B, T]
        if noisy_y.dim() == 2:
            noisy_y = noisy_y.unsqueeze(1)

        nsample = noisy_y.size(-1)
        mag_mix, phase_mix, real_mix, imag_mix = self.stft(noisy_y)  # [B=2, C=1, F=1025, T=87]
        batch_size, num_channels, n_frame = noisy_y.shape

        # Sub-band-wise RNN
        spec_RI = torch.cat([real_mix, imag_mix], 1)  # B, C * ri, F, T
        spec = torch.complex(real_mix, imag_mix)  # B, C, F, T
        subband_spec_RI = []
        subband_spec_complex = []
        band_idx = 0
        for i in range(len(self.band_width)):
            # B, C * ri, Fs, T
            subband_spec_ri_i = spec_RI[:, :, band_idx : band_idx + self.band_width[i]].contiguous()
            # B, Fs, T
            subband_spec_complex_i = spec[:, 0, band_idx : band_idx + self.band_width[i]].contiguous()
            subband_spec_RI.append(subband_spec_ri_i)
            subband_spec_complex.append(subband_spec_complex_i)
            band_idx += self.band_width[i]

        subband_feat = []
        for i in range(len(self.band_width)):
            subband_spec_RI_i = rearrange(subband_spec_RI[i], "b c fs t -> b (c fs) t")  # [B, ri * C * Fs, T]
            subband_feat.append(self.BN[i](subband_spec_RI_i))  # [B, e, T]
        subband_feat = torch.stack(subband_feat, 1)  # [B, n_band, e, T],

        # separator
        # [B, n_band*N, T]
        sep_output = self.separator([rearrange(subband_feat, "b n e t -> b (n e) t"), self.n_band, 0])  # type: ignore

        if isinstance(sep_output, list):
            sep_output = sep_output[0]
        sep_output = sep_output.view(batch_size, self.n_band, self.args.feat_dim, -1)

        spk_1_sep_subband_spec = []
        spk_2_sep_subband_spec = []
        for i in range(len(self.band_width)):
            this_output = self.mask[i](sep_output[:, i]).view(batch_size, 2, 2, 2, self.band_width[i], -1)

            this_mask = this_output[:, 0] * torch.sigmoid(this_output[:, 1])  # B*nch, 2, 2, BW, T
            this_mask_real = this_mask[:, 0]  # B*nch, BW, T
            this_mask_imag = this_mask[:, 1]  # B*nch, BW, T

            spk_1_this_mask_real, spk_2_this_mask_real = this_mask_real[:, 0], this_mask_real[:, 1]
            spk_1_this_mask_imag, spk_2_this_mask_imag = this_mask_imag[:, 0], this_mask_imag[:, 1]

            spk_1_est_spec_real = (
                subband_spec_complex[i].real * spk_1_this_mask_real
                - subband_spec_complex[i].imag * spk_1_this_mask_imag
            )  # B*nch, BW, T
            spk_1_est_spec_imag = (
                subband_spec_complex[i].real * spk_1_this_mask_imag
                + subband_spec_complex[i].imag * spk_1_this_mask_real
            )  # B*nch, BW, T

            spk_2_est_spec_real = (
                subband_spec_complex[i].real * spk_2_this_mask_real
                - subband_spec_complex[i].imag * spk_2_this_mask_imag
            )
            spk_2_est_spec_imag = (
                subband_spec_complex[i].real * spk_2_this_mask_imag
                + subband_spec_complex[i].imag * spk_2_this_mask_real
            )

            spk_1_sep_subband_spec.append(torch.complex(spk_1_est_spec_real, spk_1_est_spec_imag))
            spk_2_sep_subband_spec.append(torch.complex(spk_2_est_spec_real, spk_2_est_spec_imag))

        spk_1_est_spec = torch.cat(spk_1_sep_subband_spec, 1)  # B*nch, F, T
        if spec.shape[1] > spk_1_est_spec.shape[1]:
            spk_1_est_spec = torch.cat([spk_1_est_spec, spec[:, spk_1_est_spec.shape[1] :, :]], 1)

        spk_2_est_spec = torch.cat(spk_2_sep_subband_spec, 1)
        if spec.shape[1] > spk_2_est_spec.shape[1]:
            spk_2_est_spec = torch.cat([spk_2_est_spec, spec[:, spk_2_est_spec.shape[1] :, :]], 1)

        output_1 = self.istft(spk_1_est_spec, length=nsample, input_type="complex")
        output_2 = self.istft(spk_2_est_spec, length=nsample, input_type="complex")

        output = torch.stack([output_1, output_2], dim=1)

        return output


if __name__ == "__main__":
    batch_size = 2
    num_channels = 1  # 单通道输入信号
    num_samples = 44100  # 1s 采样点数，采样率 44100
    mixture = torch.rand(batch_size, num_channels, num_samples)
    args = ModelArgs()
    model = Model(args)
    output = model(mixture)
    print(output.shape)

    # output: torch.Size([2, 2, 44100])
    # 2: batch_size
    # 2: num_sources
    # 44100: num_samples



import logging
import math
import shutil
import time
from functools import partial
from pathlib import Path
from typing import Any, Dict, Optional, Tuple, Type, Union

import librosa
import numpy as np
import pandas as pd
import torch
import torch.distributed as dist
import torch.nn as nn
from torch.nn.parallel import DistributedDataParallel as DDP
from torch.utils.data import DataLoader, Dataset, DistributedSampler
from torchinfo import summary
from tqdm import tqdm

from audiozen.accelerate import gather_object, get_world_size_and_rank, wait_for_everyone
from audiozen.logger import TensorboardLogger
from audiozen.optimization import (
    get_constant_schedule_with_warmup,
    get_cosine_with_min_lr_schedule_with_warmup,
    get_linear_schedule_with_warmup,
)
from audiozen.trainer_args import TrainingArgs
from audiozen.trainer_utils import (
    StateName,
    TrainerState,
    collect_env,
    load_state_dict_from_checkpoint,
    print_commandline_args,
    save_state_dict_to_checkpoint,
)
from audiozen.utils import cleanup_before_training, prepare_empty_dir, set_random_seed


logger = logging.getLogger(__name__)


class Trainer:
    def __init__(
        self,
        model: nn.Module,
        args: TrainingArgs,
        data_collator: Optional[Any] = None,
        train_dataset: Optional[Dataset] = None,
        eval_dataset: Optional[Union[Dataset, Dict[str, Dataset]]] = None,
        optimizers: Tuple[Optional[torch.optim.Optimizer], Optional[torch.optim.lr_scheduler.LambdaLR]] = (None, None),
        **kwargs,
    ) -> None:
        """Initialize the Trainer.

        Args:
            model: The model to train, evaluate, or test.
            args: The training arguments. This is a dataclass that contains all the training arguments.
            data_collator: The data collator to use. Defaults to None.
            train_dataset: The training dataset. Defaults to None.
            eval_dataset: The evaluation dataset. Defaults to None. If can be a single dataset or a dictionary of datasets.
            optimizers: The optimizer and the learning rate scheduler. Defaults to (None, None).

        Important attributes:
            **in_training**: Whether or not a model is currently running `train` (e.g. when `evaluate` is called while in `train`)
        """
        self.in_training = False
        # Set a flag tensor for early stopping and other breakpoints
        self.flag_tensor = None

        self.args = args

        # Set device as early as possible
        self.world_size, self.rank = get_world_size_and_rank()
        self.is_rank_zero = self.rank == 0

        # Set random seed
        self.seed = set_random_seed(seed=args.seed)

        # Datasets, model, optimizer, lr_scheduler
        self.data_collator = data_collator
        self.train_dataset = train_dataset

        # Dataloaders. They will be created in the `train` and `evaluate` methods.
        self.train_dataloader = None
        self.eval_dataloaders = None

        self.eval_dataset = eval_dataset
        self.model = model
        self.optimizer, self.lr_scheduler = optimizers

        # Setup directories
        self._setup_exp_paths(output_dir=args.output_dir)

        # Acoustic parameters
        self._inject_acoustic_args()

        # Trainer states
        self.state = TrainerState(greater_is_better=self.args.greater_is_better)

        # Pandas settings for better display
        pd.set_option("display.float_format", lambda x: "%.3f" % x)

        if self.is_rank_zero:
            prepare_empty_dir(
                [self.checkpoints_dir, self.tb_log_dir, self.enhanced_dir, self.metrics_dir],
                resume=self.args.resume_from_checkpoint != "no",
            )

            logger.info(f"\nEnvironment information:\n{collect_env()}")

            logger.info("===== Training Arguments =====")
            logger.info(print_commandline_args())
            logger.info("=" * 30)

            self.writer = TensorboardLogger(self.tb_log_dir.as_posix())
            self.writer.log_config(self.args.to_dict())

            # Model summary
            logger.info(f"\n{summary(self.model, verbose=0)}")

        wait_for_everyone()

    @staticmethod
    def _search_latest_ckpt_path(checkpoints_dir: Path) -> Path:
        """Find the latest checkpoint path from a given checkpoint directory."""
        # Pick up all checkpoints with the format `epoch_*`
        checkpoints = sorted(checkpoints_dir.glob("epoch_" + ("[0-9]" * 4)))

        # Remove files that is not a checkpoint
        checkpoints = [ckpt for ckpt in checkpoints if ckpt.is_dir()]

        if len(checkpoints) == 0:
            raise FileNotFoundError(f"No checkpoints found in {checkpoints_dir.as_posix()}.")

        # Pick up the latest checkpoint
        ckpt_path = checkpoints[-1]

        return ckpt_path

    def _parse_ckpt_path_or_alias(self, path_or_alias: str) -> Optional[Path]:
        """Parse the checkpoint path or alias.

        If the path_or_alias is "no", it means that the training starts from scratch. It returns None.
        """
        if path_or_alias not in ["no", "latest", "best"] and not Path(path_or_alias).exists():
            raise FileNotFoundError(f"Checkpoint not found in {path_or_alias}")

        ckpt_path = None

        if path_or_alias == "no":
            if self.is_rank_zero:
                logger.info("Training from scratch. No checkpoint is loaded.")
            return ckpt_path
        elif path_or_alias == "latest":
            ckpt_path = self._search_latest_ckpt_path(self.checkpoints_dir)
        elif path_or_alias == "best":
            ckpt_path = self.checkpoints_dir / "best"
        else:
            ckpt_path = Path(path_or_alias).expanduser().resolve()

        if not ckpt_path.exists():
            raise FileNotFoundError(f"Checkpoint not found in {ckpt_path.as_posix()}")

        return ckpt_path

    def _load_state_dict_from_checkpoint(self, path_or_alias: str) -> Dict[str, Any]:
        """Load the state dictionary from the checkpoint directory."""
        if path_or_alias not in ["no", "latest", "best"] and not Path(path_or_alias).exists():
            raise FileNotFoundError(f"Checkpoint not found in {path_or_alias}")

        state_dict = {}

        # Parse the checkpoint path or alias
        if path_or_alias == "no":
            if self.is_rank_zero:
                logger.info("Training from scratch. No checkpoint is loaded.")
            return state_dict
        elif path_or_alias == "latest":
            ckpt_path = self._search_latest_ckpt_path(self.checkpoints_dir)
        elif path_or_alias == "best":
            ckpt_path = self.checkpoints_dir / "best"
        else:
            ckpt_path = Path(path_or_alias).expanduser().resolve()

        if not ckpt_path.exists():
            raise FileNotFoundError(f"Checkpoint not found in {ckpt_path.as_posix()}")

        # Load the checkpoint
        return load_state_dict_from_checkpoint(ckpt_path)

    def save_checkpoint(self, epoch: int, is_best_epoch: bool = False):
        """Save the checkpoint.

        Args:
            epoch: The current epoch.
            is_best_epoch: Whether the current epoch is the best epoch. Defaults to False.
        """

        if is_best_epoch:
            ckpt_path = self.checkpoints_dir / "best"
        else:
            ckpt_path = self.checkpoints_dir / f"epoch_{str(epoch).zfill(4)}"

        if isinstance(self.model, DDP):
            logger.info("Extract the `state_dict` from the `model.module.state_dict()`...")
            model_stain_dict = self.model.module.state_dict()
        else:
            logger.info("Extract the `state_dict` from the `model.state_dict()`...")
            model_stain_dict = self.model.state_dict()

        state_dict = {StateName.MODEL: model_stain_dict}
        state_dict[StateName.TRAINER_STATE] = self.state.state_dict()
        state_dict[StateName.OPTIMIZER] = self.optimizer.state_dict() if self.optimizer is not None else None
        state_dict[StateName.LR_SCHEDULER] = self.lr_scheduler.state_dict() if self.lr_scheduler is not None else None

        # Find all checkpoints
        checkpoints = sorted(self.checkpoints_dir.glob("epoch_" + ("[0-9]" * 4)))
        if epoch < len(checkpoints):
            logger.warning(
                f"Current epoch is {epoch}, but found {len(checkpoints)} checkpoints. "
                f"This may be caused by you running the same experiment multiple times. "
                f"Recommend to run the experiment with a different `output_dir`."
                f"Otherwise, the newer checkpoints will be removed."
            )

        # Remove the old checkpoints
        if len(checkpoints) > self.args.save_total_limit:
            logger.info(
                f"Found {len(checkpoints)} checkpoints, keeping the latest {self.args.save_total_limit} checkpoints."
            )

            for checkpoint_dir in checkpoints[: -self.args.save_total_limit]:
                shutil.rmtree(checkpoint_dir.as_posix())
                logger.info(f"Checkpoint {checkpoint_dir.as_posix()} is removed.")

        save_state_dict_to_checkpoint(ckpt_path, state_dict)

    def _prepare_ddp_model(self, model):
        """Prepare the model for acceleration.

        Notes:
            You should never try to change your model's parameters after wrapping up your model with DDP. Because,
            when wrapping up your model with DDP, the constructor of DDP will register the additional gradient reduction
            functions on all the parameters of the model itself at time of construction. If you change the parameters of
            the model after wrapping up the model with DDP, the gradient reduction functions will not be matched with the
            correct parameters of the model. It means that you must load the model's weights before wrapping up the model
            with DDP. See: https://pytorch.org/docs/stable/generated/torch.nn.parallel.DistributedDataParallel.html

        Args:
            model: The model to prepare.

        Returns:
            The model prepared for acceleration.
        """
        if dist.is_available() and dist.is_initialized():
            model = model.to(self.rank)
            model = DDP(model, find_unused_parameters=self.args.ddp_find_unused_parameters)
            # synchronize before training begins
            dist.barrier()
            return model
        else:
            model = model.to("cpu")
            return model

    @staticmethod
    def _current_time_now():
        return time.strftime("%Y_%m_%d--%H_%M_%S")

    def _setup_exp_paths(self, output_dir: str):
        """Set the paths for the experiment.

        Args:
            output_dir: the root directory to save all experiments.

        Example:
            - output_dir: /home/xhao/exp/fullsubnet_lr_0.g
            - checkpoints_dir: /home/xhao/exp/fullsubnet_lr_0.1/checkpoints
            - tb_log_dir: /home/xhao/exp/fullsubnet_lr_0.1/tb_log
            - enhanced_dir: /home/xhao/exp/fullsubnet_lr_0.1/enhanced
            - metrics_dir: /home/xhao/exp/fullsubnet_lr_0.1/metrics
            - source_code_backup_dir: /home/xhao/exp/fullsubnet_lr_0.1/source_code__YYYY_MM_DD__HH_MM_SS
            - model_args_path: /home/xhao/exp/fullsubnet_lr_0.1/model_args__YYYY_MM_DD__HH_MM_SS.yaml
        """
        time_now = self._current_time_now()  # returns a timestamp string

        self.output_dir = Path(output_dir).expanduser().absolute()
        self.checkpoints_dir = self.output_dir / "checkpoints"
        self.tb_log_dir = self.output_dir / "tb_log"
        self.enhanced_dir = self.output_dir / "enhanced"
        self.metrics_dir = self.output_dir / "metrics"

        # Each run will have a unique source code, config, and log file.
        self.source_code_dir = Path(__file__).expanduser().absolute().parent.parent.parent
        self.source_code_backup_dir = self.output_dir / f"source_code__{time_now}"
        self.model_args_path = self.output_dir / f"model_args__{time_now}.yaml"
        self.loss_log_path = self.output_dir / "loss.csv"

    def _inject_acoustic_args(self):
        """Setup acoustic arguments."""
        n_fft = self.args.acoustic_n_fft
        hop_length = self.args.acoustic_hop_length
        win_length = self.args.acoustic_win_length
        sr = self.args.acoustic_sr

        # Support for torch and librosa stft
        self.torch_stft = partial(torch.stft, n_fft=n_fft, hop_length=hop_length, win_length=win_length)
        self.torch_istft = partial(torch.istft, n_fft=n_fft, hop_length=hop_length, win_length=win_length)
        self.librosa_stft = partial(librosa.stft, n_fft=n_fft, hop_length=hop_length, win_length=win_length)
        self.librosa_istft = partial(librosa.istft, n_fft=n_fft, hop_length=hop_length, win_length=win_length)

        self.n_fft = n_fft
        self.hop_length = hop_length
        self.win_length = win_length
        self.sr = sr

    def set_models_to_train_mode(self):
        """Set models to train mode.

        You can override this method to set your own models to train mode. For example, in GAN training, you may want
        to set the generator and the discriminator to train mode.
        """
        self.model.train()

    def set_models_to_eval_mode(self):
        self.model.eval()

    def lr_scheduler_step(self):
        """Step the lr scheduler.

        You can override this method to step your own lr scheduler. For example, in GAN training, you may want to step
        the lr scheduler of the generator and the discriminator.
        """
        assert self.lr_scheduler is not None, "lr_scheduler is not initialized."
        self.lr_scheduler.step()

    def create_bar_desc(self, loss_dict: Dict[str, float]):
        """Create the description for the progress bar."""
        assert self.lr_scheduler is not None, "lr_scheduler is not initialized."

        bar_desc = ""
        for k, v in loss_dict.items():
            bar_desc += f"{k}: {(v):.4f}, "
        bar_desc += f"lr: {self.lr_scheduler.get_last_lr()[-1]:.10f}"

        if self.args.plot_lr:
            self.writer.add_scalar("Train_Step/lr", self.lr_scheduler.get_last_lr()[-1], self.state.steps_trained)

        return bar_desc

    def _has_improved(self, score, save_max_score=True):
        """Check if the current model got the best metric score."""
        if save_max_score:
            return score > self.state.best_score
        else:
            return score < self.state.best_score

    def _should_early_stop(self, score: float):
        should_stop = False

        if self._has_improved(score, save_max_score=self.args.greater_is_better):
            self.state.best_score = score
            self.state.best_score_epoch = self.state.epochs_trained
            self.save_checkpoint(self.state.epochs_trained, is_best_epoch=True)
            self.state.early_stopping_patience_counter = 0
            logger.info(f"Found new best score: {score:.4f}, saving checkpoint...")
        else:
            logger.info(
                f"Score did not improve from {self.state.best_score:.4f} at epoch {self.state.best_score_epoch}."
            )
            self.state.early_stopping_patience_counter += 1
            logger.info(
                f"Early stopping counter: {self.state.early_stopping_patience_counter} out of {self.args.early_stopping_patience}"
            )

            if self.state.early_stopping_patience_counter >= self.args.early_stopping_patience:
                logger.info("Early stopping triggered, stopping training...")
                should_stop = True

        return should_stop

    def init_train_dataloader(self):
        if self.train_dataset is None:
            raise ValueError("Trainer: `train()` requires a `train_dataset`.")

        if dist.is_available() and dist.is_initialized():
            sampler = DistributedSampler(
                self.train_dataset,
                num_replicas=self.world_size,
                rank=self.rank,
                shuffle=True,
                seed=0,  # This number should be identical across all processes in the distributed group
            )
        else:
            sampler = None

        dataloader_params = {
            "batch_size": self.args.per_device_train_batch_size,
            "collate_fn": self.data_collator,
            "num_workers": self.args.dataloader_num_workers,
            "pin_memory": self.args.dataloader_pin_memory,
            "persistent_workers": self.args.dataloader_persistent_workers,
            "drop_last": self.args.dataloader_drop_last,
            "prefetch_factor": self.args.dataloader_prefetch_factor,
            "sampler": sampler,
            "shuffle": (sampler is None),
        }

        if self.is_rank_zero:
            logger.info("Dataset and Sampler are initialized.")

        return DataLoader(self.train_dataset, **dataloader_params)

    def init_eval_dataloaders(self) -> Dict[str, DataLoader]:
        """Create the evaluation dataloaders.

        If the eval_dataset is a single dataset, it will be converted to a dictionary with the key "default".

        Returns:
            eval_dataloaders: the evaluation dataloaders.
        """
        if self.eval_dataset is None:
            raise ValueError("Trainer: evaluation requires a `eval_dataset`.")

        eval_dataset = self.eval_dataset
        data_collator = self.data_collator

        if not isinstance(eval_dataset, dict):
            if isinstance(eval_dataset, Dataset):
                eval_dataset = {"default": eval_dataset}
            else:
                raise ValueError("Trainer: `eval_dataset` should be either a dataset or a dictionary of datasets.")

        dataloader_params = {
            "batch_size": self.args.eval_batch_size,
            "collate_fn": data_collator,
            "num_workers": self.args.dataloader_num_workers,
            "pin_memory": self.args.dataloader_pin_memory,
            "persistent_workers": self.args.dataloader_persistent_workers,
            "drop_last": False,
            "prefetch_factor": self.args.dataloader_prefetch_factor,
            "shuffle": False,  # No need to shuffle for evaluation
        }

        eval_dataloaders = {}
        for key, dataset in eval_dataset.items():
            if dist.is_available() and dist.is_initialized():
                sampler = DistributedSampler(
                    dataset,
                    num_replicas=self.world_size,
                    rank=self.rank,
                    shuffle=False,  # No need to shuffle for evaluation
                    seed=0,  # This number should be identical across all processes in the distributed group
                )
            else:
                sampler = None

            eval_dataloaders[key] = DataLoader(
                dataset=dataset,
                sampler=sampler,
                **dataloader_params,
            )

        if self.is_rank_zero:
            logger.info("Evaluation dataloaders are initialized.")
            logger.info(f"Number of evaluation dataloaders: {len(eval_dataloaders)}")

        return eval_dataloaders

    @staticmethod
    def get_optimizer_cls_and_kwargs(args: TrainingArgs) -> Tuple[Type[torch.optim.Optimizer], Dict[str, Any]]:
        """Returns the optimizer class and optimizer parameters based on the training arguments."""
        optimizer_kwargs = {"lr": args.learning_rate}

        adam_kwargs = {"betas": (args.adam_beta1, args.adam_beta2), "eps": args.adam_epsilon}

        if args.optim == "adamw":
            optimizer_cls = torch.optim.AdamW
            optimizer_kwargs |= adam_kwargs
        elif args.optim == "adam":
            optimizer_cls = torch.optim.Adam
            optimizer_kwargs |= adam_kwargs
        else:
            raise ValueError(f"Unknown optimizer: {args.optim}")

        return optimizer_cls, optimizer_kwargs

    def init_optimizer(self):
        if self.optimizer is None:
            optimizer_cls, optimizer_kwargs = self.get_optimizer_cls_and_kwargs(self.args)
            self.optimizer = optimizer_cls(self.model.parameters(), **optimizer_kwargs)

    def get_warmup_steps(self, warmup_steps, max_steps, warmup_ratio):
        """Calculate the number of warmup steps based on the training arguments."""
        if warmup_steps > 0:
            if self.is_rank_zero:
                logger.info(f"warmup_steps={warmup_steps}. warmup_ratio will be ignored.")
            return warmup_steps
        else:
            return math.ceil(max_steps * warmup_ratio)

    def create_warmup_scheduler(self, optimizer, scheduler_name, max_steps: int):
        num_warmup_steps = self.get_warmup_steps(self.args.warmup_steps, max_steps, self.args.warmup_ratio)

        if scheduler_name == "constant_schedule_with_warmup":
            return get_constant_schedule_with_warmup(optimizer=optimizer, num_warmup_steps=num_warmup_steps)
        elif scheduler_name == "linear_schedule_with_warmup":
            return get_linear_schedule_with_warmup(
                optimizer=optimizer, num_warmup_steps=num_warmup_steps, num_training_steps=max_steps
            )
        elif scheduler_name == "cosine_with_min_lr_schedule_with_warmup":
            return get_cosine_with_min_lr_schedule_with_warmup(
                optimizer=optimizer,
                num_warmup_steps=num_warmup_steps,
                num_training_steps=max_steps,
                num_cycles=self.args.cosine_with_min_lr_schedule_with_warmup_num_cycles,
                initial_lr=self.args.cosine_with_min_lr_schedule_with_warmup_initial_lr,
                min_lr=self.args.cosine_with_min_lr_schedule_with_warmup_min_lr,
            )
        else:
            raise ValueError(f"Invalid scheduler name: {scheduler_name}")

    def init_lr_scheduler(self, max_steps: int, scheduler_specific_kwargs: Optional[Dict] = None):
        """Setup the learning rate scheduler.

        You can override this method to create your own schedulers. For example, in GAN training, you may want to
        create two schedulers for the generator and the discriminator.

        Args:
            max_steps: the maximum number of steps to train.
        """
        self.lr_scheduler = self.create_warmup_scheduler(
            optimizer=self.optimizer, scheduler_name=self.args.lr_scheduler_type, max_steps=max_steps
        )

    def setup_optimizer_and_scheduler(self, max_steps: int):
        """Setup the optimizer and the learning rate scheduler."""
        self.init_optimizer()
        self.init_lr_scheduler(max_steps=max_steps)

    def set_trigger(self):
        """Sets the internal trigger tensor to 1 on the current process. A latter check should follow using this which
        will check across all processes.

        Notes:
            Does not require `wait_for_everyone()`
        """
        if dist.is_available() and dist.is_initialized():
            self.flag_tensor = torch.tensor(1, device=self.rank)

    def check_trigger(self):
        """Checks if the internal trigger tensor has been set to 1 in any of the processes. If so, will return `True`
        and reset the trigger tensor to 0.

        Notes:
            Requires `wait_for_everyone()`
        """
        if dist.is_available() and dist.is_initialized():
            # Make sure that all processes have set the trigger tensor
            # Now that we are outside `__init__`, we can initialize it if it is `None` on device
            if self.flag_tensor is None:
                self.flag_tensor = torch.tensor(0, device=self.rank)

            dist.all_reduce(self.flag_tensor, op=dist.ReduceOp.SUM)

            if self.flag_tensor.item() >= 1:
                self.flag_tensor = torch.tensor(0, device=self.rank)
                return True

            return False

    def train(self):
        """Main training entry point."""
        self.in_training = True

        cleanup_before_training()

        # Initialize the training dataloader
        train_dataloader = self.init_train_dataloader()

        # Initialize the training variables
        num_update_steps_per_epoch = len(train_dataloader)  # #samples // batch_size // world_size
        num_update_steps_per_epoch = num_update_steps_per_epoch // self.args.gradient_accumulation_steps
        num_update_steps_per_epoch = max(num_update_steps_per_epoch, 1)
        if self.args.max_steps > 0:
            # max_steps is set, so we need to calculate the number of epochs based on max_steps
            max_steps = self.args.max_steps
            num_train_epochs = max_steps // num_update_steps_per_epoch + int(
                max_steps % num_update_steps_per_epoch > 0
            )
        else:
            num_train_epochs = self.args.num_train_epochs
            max_steps = math.ceil(num_train_epochs * num_update_steps_per_epoch)

        # Load checkpoint before setting up the model
        ckpt_path = self._parse_ckpt_path_or_alias(self.args.resume_from_checkpoint)
        state_dict = None
        if ckpt_path is not None:
            state_dict = load_state_dict_from_checkpoint(ckpt_path)
            self.model.load_state_dict(state_dict[StateName.MODEL])

        # Wrappering the model for DDP
        self.model = self._prepare_ddp_model(self.model)
        # Create optimizer and lr_scheduler
        self.setup_optimizer_and_scheduler(max_steps=max_steps)

        if ckpt_path is not None and state_dict is not None:
            # Make sure that the optimizer and lr_scheduler are created after the model is wrapped up with DDP
            assert self.optimizer is not None, "Optimizer is not initialized."
            assert self.lr_scheduler is not None, "lr_scheduler is not initialized."
            self.optimizer.load_state_dict(state_dict[StateName.OPTIMIZER])
            self.lr_scheduler.load_state_dict(state_dict[StateName.LR_SCHEDULER])
            self.state.load_state_dict(state_dict[StateName.TRAINER_STATE])

        # Train!!!
        if self.is_rank_zero:
            logger.info("***** Running training *****")
            logger.info(f"  Num Epochs = {num_train_epochs:,}")
            logger.info(f"  `steps_per_epoch` = {num_update_steps_per_epoch:,}")
            logger.info(f"  Instantaneous batch size per device = {self.args.per_device_train_batch_size:,}")
            logger.info(f"  Gradient Accumulation steps = {self.args.gradient_accumulation_steps}")
            logger.info(f"  Total optimization steps = {max_steps:,}")

        for epoch in range(self.state.epochs_trained + 1, num_train_epochs + 1):
            if self.is_rank_zero:
                logger.info(f"{'=' * 9} Epoch {epoch} out of {num_train_epochs} {'=' * 9}")
                logger.info("Begin training...")

            if dist.is_available() and dist.is_initialized():
                train_dataloader.sampler.set_epoch(epoch)  # type: ignore

            self.set_models_to_train_mode()

            training_epoch_output = []

            # The iter number of progress bar increments by 1 by default whether gradient accumulation is used or not.
            # but we update the description of the progress bar only when the gradients are synchronized across all processes.
            dataloader_bar = tqdm(
                train_dataloader,
                desc="",
                dynamic_ncols=True,
                bar_format="{l_bar}{r_bar}",
                colour="green",
                disable=not self.is_rank_zero,
                position=0,
                leave=True,
            )

            for batch_idx, batch in enumerate(dataloader_bar):
                # You are responsible for calling `.backward()`, `.step()`, and `.zero_grad()` in your implementation
                loss_dict = self.training_step(batch, batch_idx)
                training_epoch_output.append(loss_dict)
                if self.is_rank_zero:
                    bar_desc = self.create_bar_desc(loss_dict)
                    dataloader_bar.set_description_str(bar_desc)

                self.lr_scheduler_step()
                self.state.steps_trained += 1
            self.state.epochs_trained += 1

            # Clean cache
            # [Not necessary in most cases]
            # Normally, PyTorch will occupy maximum GPU memory after training for a while. This will prevent other
            # processes from other users from using the GPU. It is a safe practice. Tt keeps it in a pool so that next
            # allocations can be done much faster.
            # However, we need to use more memeroy during evaluation (DNSMOS), so we need to clean the cache.
            torch.cuda.empty_cache()

            # Hook `training_epoch_end`
            self.training_epoch_end(training_epoch_output)

            # Should save?
            if self.is_rank_zero and epoch % self.args.save_epoch_interval == 0:
                self.save_checkpoint(epoch=epoch, is_best_epoch=False)

            # Should evaluate?
            if epoch % self.args.eval_epoch_interval == 0:
                if self.is_rank_zero:
                    logger.info("Training finished, begin evaluation...")

                with torch.no_grad():
                    # only the main process will receive the score
                    score = self.evaluate()
                    if self.is_rank_zero:
                        should_stop = self._should_early_stop(score)
                        if should_stop:
                            self.set_trigger()

                if self.is_rank_zero:
                    logger.info("Evaluation finished.")

            wait_for_everyone()
            # Check if any process has set the trigger tensor
            if self.check_trigger():
                break

    @torch.no_grad()
    def evaluate(self):
        """Run evaluation (validation) and returns metrics.

        Returns:
            score: The representative score of the evaluation.
        """
        if self.is_rank_zero:
            logger.info("Begin evaluation...")

        # init_eval_dataloaders() might be called multiple times if the evaluation is called during training.
        # Here, we will make sure that the dataloaders are only prepared once.
        if self.eval_dataloaders is None:
            if self.is_rank_zero:
                logger.info("The evaluation dataloaders are not initialized. Initialize them now...")
            self.eval_dataloaders = self.init_eval_dataloaders()

        # If the current run is not in training (use train mode), we need to prepare the model for DDP.
        if not self.in_training:
            # Load checkpoint before setting up the model
            ckpt_path = self._parse_ckpt_path_or_alias(self.args.resume_from_checkpoint)

            if ckpt_path is not None:
                if self.is_rank_zero:
                    logger.info(f"Load the checkpoint from {ckpt_path.as_posix()}...")

                state_dict = load_state_dict_from_checkpoint(ckpt_path)
                # Load the model state before wrapping up the model with DDP according to the PyTorch documentation
                self.model.load_state_dict(state_dict[StateName.MODEL])

            # Prepare the model for DDP
            if self.is_rank_zero:
                logger.info("Prepare the model for DDP...")

            self.model = self._prepare_ddp_model(self.model)

        evaluation_output = self.evaluation_loop(description="evaluate", gather_step_output=True)

        if self.is_rank_zero:
            # only the main process will run evaluation_epoch_end
            logger.info("Evaluation finished, begin hook `evaluation_epoch_end`...")
            score = self.evaluation_epoch_end(evaluation_output)
            return score
        else:
            return 0.0

    @torch.no_grad()
    def predict(self):
        """Run prediction.

        In the predict mode, the model will be loaded from a checkpoint and the evaluation_loop will be called.
        However, the evaluation_loop will not gather any step_output from all processes.
        """
        if self.is_rank_zero:
            logger.info("Begin predicting...")

        # In the predict mode, get_eval_dataloaders() will be called only once.
        self.eval_dataloaders = self.init_eval_dataloaders()

        # Load checkpoint before setting up the model
        ckpt_path = self._parse_ckpt_path_or_alias(self.args.resume_from_checkpoint)
        state_dict = None
        if ckpt_path is not None:
            state_dict = load_state_dict_from_checkpoint(ckpt_path)

        # Load the model state before wrapping up the model with DDP according to the PyTorch documentation
        assert state_dict is not None and StateName.MODEL in state_dict, "The model state is not found in the ckpt."
        self.model.load_state_dict(state_dict[StateName.MODEL])
        self.model = self._prepare_ddp_model(self.model)

        # In the predict mode, we don't need to gather the step_output from all processes.
        self.evaluation_loop(description="predict", gather_step_output=False)

        logger.info("Prediction finished.")

    @torch.no_grad()
    def evaluation_loop(self, description: str, gather_step_output: bool = False):
        """Prediction/evaluation loop, shared by `Trainer.evaluate()` and `Trainer.predict()`."""
        args = self.args

        self.set_models_to_eval_mode()

        if self.is_rank_zero:
            logger.info(f"***** Running {description} *****")
            logger.info(f"  Batch size = {args.eval_batch_size}")

        evaluation_output = {}
        assert self.eval_dataloaders is not None, "The evaluation dataloaders are not initialized."
        for dl_idx, (dl_id, dataloader) in enumerate(self.eval_dataloaders.items()):
            dataloader_output = []
            for batch_idx, batch in enumerate(
                tqdm(
                    dataloader,
                    desc=f"Evaluation on dataloader `{dl_id}`",
                    bar_format="{l_bar}{r_bar}",
                    dynamic_ncols=True,
                    disable=not self.is_rank_zero,
                )
            ):
                """
                It is advised against computing metrics within the `evaluation_epoch_end` method for several reasons:
                    1. Most evaluation metrics are inherently sequential and not parallelizable. Hence, computing them in `evaluation_epoch_end` does not offer a speed advantage during evaluation.
                    2. By not aggregating all outputs for metric calculation at the epoch's end, we reduce the risk of memory overflow, which can occur when gathering results across all processes.
                    3. Calculating the metric score during `evaluation_step` allows for earlier detection of any errors in the code.

                Recommendations for metric calculation:
                    1. Perform immediate metric score calculation within the `evaluation_step` method.
                    2. Accumulate the results at this stage.
                    3. If necessary, compute the average or aggregate metric score in the `evaluation_epoch_end` method.
                """
                with torch.no_grad():
                    step_output = self.evaluation_step(batch, batch_idx, dl_id)

                # If `gather_step_output` is True, we will gather the step_output from all processes and return a list of all metric scores.
                if gather_step_output:
                    """Collect the step_output from all processes and return a list of all metric scores.

                    Assume we have two processes:
                    step_output = [
                        {"metric_1": xx, "metric_2": xx, ...},  # process 0
                        {"metric_1": xx, "metric_2": xx, ...},  # process 1
                        {"metric_1": xx, "metric_2": xx, ...},  # process 0
                        {"metric_1": xx, "metric_2": xx, ...},  # process 1
                        ...
                    ]
                    """
                    step_output = gather_object(step_output, dst=0)
                dataloader_output.append(step_output)
            evaluation_output[dl_id] = dataloader_output

        """
        evaluation_output = {
            "dataloader_id_1": [step_output_0, step_output_1, ...],
            "dataloader_id_2": [step_output_0, step_output_1, ...],
            ...
        }
        """
        return evaluation_output

    def training_step(self, batch, batch_idx):
        """Implement a training step (iteration).

        Implement your own training step here. The input batch is from a training dataloader and the output of this
        function can be various. For example, it can be a dict of loss, or a dict of loss and some enhanced audio signals.
        Here is the persuade code for training a model:

        .. code-block:: python
            :emphasize-lines: 6

            for epoch in range(start_epoch, end_epoch):
                self.model.train()

                training_epoch_output = []
                for batch, batch_index in dataloader:
                    loss_dict = training_step(batch, batch_idx)
                    training_epoch_output.append(loss_dict)

                training_epoch_end(training_epoch_output)
                save_checkpoint()
                if some_condition:
                    score = validate()
                    if score > best_score:
                        save_checkpoint(best=True)

        Args:
            batch: a batch of data, which passed from a custom training dataloader.
            batch_idx: the index of the current batch.

        Returns:
            loss_dict: a dict of loss. For example, {"loss_1": loss, "loss_2": loss, ...}
        """
        raise NotImplementedError

    def training_epoch_end(self, training_epoch_output):
        """Implement the logic of the end of a training epoch.

        By default, this function will log the mean loss of each loss item on a training epoch. You can override this.
        When the training epoch ends, this function will be called. The input is a list of the loss dict of each step
        in a training epoch. You may want to log the epoch-level training loss here.

        .. code-block:: python
            for epoch in range(start_epoch, end_epoch):
                self.model.train()

                training_epoch_output = []
                for batch, batch_index in dataloader:
                    loss = training_step(batch, batch_idx)
                    training_epoch_output.append(loss)

                training_epoch_end(training_epoch_output)
                save_checkpoint()
                if some_condition:
                    score = validate()
                    if score > best_score:
                        save_checkpoint(best=True)

        Args:
            training_epoch_output: the output of the training epoch. It may a list of the output of each batch (iteration).
        """
        # Compute mean loss on all loss items on a epoch
        if self.is_rank_zero:
            loss_keys = training_epoch_output[0].keys()
            loss_dict = {key: np.mean([step_out[key] for step_out in training_epoch_output]) for key in loss_keys}

            for key, value in loss_dict.items():
                logger.info(f"Loss '{key}' on epoch {self.state.epochs_trained}: {value}")
                self.writer.add_scalar(f"Train_Epoch/{key}", value, self.state.epochs_trained)

            # loss_dict["epoch"] = self.state.epochs_trained
            # Append the loss to the loss log
            # df = pd.DataFrame(loss_dict, index=[0])
            # df.to_csv(self.loss_log_path, mode="a", index=False, header=not self.loss_log_path.exists())

    def evaluation_step(self, batch, batch_idx, dataloader_idx):
        """Implement a evaluation/prediction step.

        This function defines the evaluation step. The input batch is from a eval dataloader.
        Here is the persuade code for validating a model:

        .. code-block:: python
            :emphasize-lines: 4

            evaluation_output = []
            for dataloader_idx, dataloader in dataloaders:
                for batch_index, batch in dataloader:
                    loss_or_data = evaluation_step(batch, batch_idx)
                    evaluation_epoch_output.append(loss_or_data)

            score = evaluation_epoch_end(evaluation_epoch_output)
            return score

        Notes:
            **The evaluation step will be run on all processes.**

            About batch size:
            If your evaluation data have the same length, you may use a batch size larger than 1 to speed up the evaluation.
            For example, if you have 1000 samples in the evaluation set, and you have a batch size of 100, then you will
            have 10 batches in the evaluation set. However, if your data in the evaluation set has a different length, please
            use a batch size of 1. It still works for distributed evaluation. Otherwise, you will get an error.

            About distributed evaluation:
            The output of this function will be gathered across all processes. For example, if you have 4 processes, and
            you have a batch size of 1, then you will have 4 outputs from this function. The output of this function will
            be gathered across all processes. The first dimension of the result is num_processes multiplied by the first
            dimension of the input tensors. **Please make sure the first dimension of the input tensors is the batch size.**
            **The last dimension of the output will be padded to the length of the longest sample in the evaluation set.**
            It means that the output will be a tensor with the shape of [num_processes * batch_size, max_length]. If you
            calculate the metric score on the output, you should do a truncation to remove the padding. Otherwise, if you
            are using a metric that sensitive to the padding, you will get a wrong metric score. It is not easy to
            implement this truncation in the ``evaluation_epoch_end`` function. We recommend you directly calculate the metric
            score in the evaluation_step function. I guess the Accelerate team will implement a automatic truncation in the
            future. https://github.com/huggingface/accelerate/issues/226

        Args:
            batch: a batch of data.
            batch_idx: the index of the batch.
            dataloader_idx: the index of the dataloader.

        Returns:
            output: the output of the batch. It may enhanced audio signals.
        """
        raise NotImplementedError

    def evaluation_epoch_end(self, outputs, log_to_tensorboard=True):
        """Evaluation epoch end.

        The input `evaluation_epoch_output` will be a list of list. For example, if you have two dataloaders, the `evaluation_epoch_output` will be:

        .. code-block:: python

            evaluation_epoch_output = [
                [dataloader_1_batch_1_output, dataloader_1_batch_2_output, ...],
                [dataloader_2_batch_1_output, dataloader_2_batch_2_output, ...],
                ...,
            ]


        The output of this function should be a metric score, which will be used to determine whether the current model is the best model.

        .. code-block:: python
            :emphasize-lines: 7

            evaluation_output = []
            for dataloader_idx, dataloader in dataloaders:
                for batch_index, batch in dataloader:
                    loss_or_data = evaluation_step(batch, batch_idx)
                    evaluation_epoch_output.append(loss_or_data)

            score = evaluation_epoch_end(evaluation_epoch_output)
            return score

        Args:
            evaluation_epoch_output: the output of the evaluation epoch. It is a list of list.

        Returns:
            score: the metric score of the evaluation epoch.
        """
        # We use this variable to store the score for the current epoch
        score = 0.0

        for dl_id, dataloader_outputs in outputs.items():
            metric_dict_list = []
            for i, step_output in enumerate(dataloader_outputs):
                metric_dict_list += step_output

            # Use pandas to compute the mean of all metrics and save them to a csv file
            df_metrics = pd.DataFrame(metric_dict_list)
            df_metrics_mean = df_metrics.mean(numeric_only=True)
            df_metrics_mean_df = df_metrics_mean.to_frame().T  # Convert mean to a DataFrame

            time_now = self._current_time_now()
            df_metrics.to_csv(
                self.metrics_dir / f"dl_{dl_id}_epoch_{self.state.epochs_trained}_{time_now}.csv", index=False
            )
            df_metrics_mean_df.to_csv(
                self.metrics_dir / f"dl_{dl_id}_epoch_{self.state.epochs_trained}_{time_now}_mean.csv", index=False
            )

            logger.info(f"\n{df_metrics_mean_df.to_markdown()}")

            # We use the `metric_for_best_model` to compute the score.
            if self.in_training:
                try:
                    score += df_metrics_mean[self.args.metric_for_best_model]
                except KeyError:
                    logger.warning(f"Metric '{self.args.metric_for_best_model}' not found in the evaluation metrics.")
                    logger.warning("You won't be able to use this metric to determine the best model.")
                    logger.warning("Please check the return in your `evaluation_step` function.")
                    score += 0.0

                if log_to_tensorboard:
                    for metric, value in df_metrics_mean.items():
                        self.writer.add_scalar(f"metrics_{dl_id}/{metric}", value, self.state.epochs_trained)

        return score



from dataclasses import dataclass
from typing import Optional

import torch
import torch.nn as nn


@dataclass
class CustomTransformerConfig:
    vocab_size: int


class Attention(nn.Module):
    def __init__(
        self,
        embed_dim: int,
        num_heads: int,
        dropout: float = 0.0,
        is_decoder: bool = False,
        bias: bool = True,
        is_causal: bool = False,
        config: Optional[CustomTransformerConfig] = None,
    ):
        self.embed_dim = embed_dim
        self.num_heads = num_heads
        self.dropout = dropout
        self.head_dim = embed_dim // num_heads
        self.config = config

        self.scaling = self.head_dim**-0.5
        self.is_decoder = is_decoder
        self.is_causal = is_causal

        self.q_proj = nn.Linear(embed_dim, embed_dim, bias=bias)
        self.k_proj = nn.Linear(embed_dim, embed_dim, bias=bias)
        self.v_proj = nn.Linear(embed_dim, embed_dim, bias=bias)
        self.out_proj = nn.Linear(embed_dim, embed_dim, bias=bias)

    # Copied from transformers.models.bart.modeling_bart.BartAttention._shape with BART->whisper
    def _shape(self, tensor: torch.Tensor, seq_len: int, bsz: int):
        return tensor.view(bsz, seq_len, self.num_heads, self.head_dim).transpose(1, 2).contiguous()

    def forward(self, hidden_states: torch.Tensor):
        """Input shape: [batch_size, seq_len, embed_dim]"""
        bsz, tgt_len, _ = hidden_states.shape
        # get query proj
        query_states = self.q_proj(hidden_states) * self.scaling
        # self_attention
        key_states = self._shape(self.k_proj(hidden_states), -1, bsz)
        value_states = self._shape(self.v_proj(hidden_states), -1, bsz)


class FlashAttention2(nn.Module):
    pass


ATTENTION_CLASSES = {
    "eager": Attention,
    "flash_attention_2": FlashAttention2,
}


class EncoderLayer(nn.Module):
    def __init__(self, conf: CustomTransformerConfig):
        super().__init__()

    def forward(self, x):
        pass



import torch
import torch.nn as nn

from audiozen.constant import EPSILON


def cumulative_norm(input):
    """Cumulative Mag normalization.

    References:
            https://arxiv.org/pdf/1809.07454
    """
    eps = EPSILON

    # [B, C, F, T]
    batch_size, n_channels, n_freqs, n_frames = input.size()
    device = input.device
    data_type = input.dtype

    input = input.reshape(batch_size * n_channels, n_freqs, n_frames)

    step_sum = torch.sum(input, dim=1)  # [B, T]
    step_pow_sum = torch.sum(torch.square(input), dim=1)

    cumulative_sum = torch.cumsum(step_sum, dim=-1)  # [B, T]
    cumulative_pow_sum = torch.cumsum(step_pow_sum, dim=-1)  # [B, T]

    entry_count = torch.arange(n_freqs, n_freqs * n_frames + 1, n_freqs, dtype=data_type, device=device)
    entry_count = entry_count.reshape(1, n_frames)  # [1, T]
    entry_count = entry_count.expand_as(cumulative_sum)  # [1, T] => [B, T]

    cum_mean = cumulative_sum / entry_count  # B, T
    cum_var = (cumulative_pow_sum - 2 * cum_mean * cumulative_sum) / entry_count + cum_mean.pow(2)  # B, T
    cum_std = (cum_var + eps).sqrt()  # B, T

    cum_mean = cum_mean.reshape(batch_size * n_channels, 1, n_frames)
    cum_std = cum_std.reshape(batch_size * n_channels, 1, n_frames)

    x = (input - cum_mean) / cum_std
    x = x.reshape(batch_size, n_channels, n_freqs, n_frames)

    return x


class CumulativeMagSpectralNorm(nn.Module):
    def __init__(self, cumulative=False, use_mid_freq_mu=False):
        """Cumulative Mag normalization.

        Args:
            cumulative: 是否采用累积的方式计算 mu
            use_mid_freq_mu: 仅采用中心频率的 mu 来代替全局 mu

        Notes:
            先算均值再累加 等同于 先累加再算均值

        References:
            https://arxiv.org/pdf/1809.07454
        """
        super().__init__()
        self.eps = EPSILON
        self.cumulative = cumulative
        self.use_mid_freq_mu = use_mid_freq_mu

    def forward(self, input):
        assert input.ndim == 4, f"{self.__name__} only support 4D input."
        batch_size, n_channels, n_freqs, n_frames = input.size()
        device = input.device
        data_type = input.dtype

        input = input.reshape(batch_size * n_channels, n_freqs, n_frames)

        if self.use_mid_freq_mu:
            step_sum = input[:, int(n_freqs // 2 - 1), :]  # [B * C, F, T] => [B * C, T]
        else:
            step_sum = torch.mean(input, dim=1)  # [B * C, F, T] => [B * C, T]

        if self.cumulative:
            cumulative_sum = torch.cumsum(step_sum, dim=-1)  # [B, T]
            entry_count = torch.arange(1, n_frames + 1, dtype=data_type, device=device)
            entry_count = entry_count.reshape(1, n_frames)  # [1, T]
            entry_count = entry_count.expand_as(cumulative_sum)  # [1, T] => [B, T]

            mu = cumulative_sum / entry_count  # [B * C, T]
            mu = mu.reshape(batch_size * n_channels, 1, n_frames)
        else:
            mu = torch.mean(step_sum, dim=-1)  # [B * C]
            mu = mu.reshape(batch_size * n_channels, 1, 1)  # [B * C, 1, 1]

        input_normed = input / (mu + self.eps)
        input_normed = input_normed.reshape(batch_size, n_channels, n_freqs, n_frames)
        return input_normed


if __name__ == "__main__":
    a = torch.rand(2, 1, 160, 200)
    ln = CumulativeMagSpectralNorm(cumulative=False, use_mid_freq_mu=False)
    ln_1 = CumulativeMagSpectralNorm(cumulative=True, use_mid_freq_mu=False)
    ln_2 = CumulativeMagSpectralNorm(cumulative=False, use_mid_freq_mu=False)
    ln_3 = CumulativeMagSpectralNorm(cumulative=True, use_mid_freq_mu=False)
    print(ln(a).mean())
    print(ln_1(a).mean())
    print(ln_2(a).mean())
    print(ln_3(a).mean())



# Adapted from huggingface/transformers/src/transformers/debug_utils.py
import collections

import torch


class DebugUnderflowOverflow:
    """This debug class helps detect and understand where the model starts getting very large or very small, and more
    importantly `nan` or `inf` weight and activation elements.

    There are 2 working modes:

    1. Underflow/overflow detection (default)
    2. Specific batch absolute min/max tracing without detection

    Mode 1: Underflow/overflow detection

    To activate the underflow/overflow detection, initialize the object with the model :

    ```python
    debug_overflow = DebugUnderflowOverflow(model)
    ```

    then run the training as normal and if `nan` or `inf` gets detected in at least one of the weight, input or
    output elements this module will throw an exception and will print `max_frames_to_save` frames that lead to this
    event, each frame reporting

    1. the fully qualified module name plus the class name whose `forward` was run
    2. the absolute min and max value of all elements for each module weights, and the inputs and output

    For example, here is the header and the last few frames in detection report for `google/mt5-small` run in fp16 mixed precision :

    ```
    Detected inf/nan during batch_number=0
    Last 21 forward frames:
    abs min  abs max  metadata
    [...]
                      encoder.block.2.layer.1.DenseReluDense.wi_0 Linear
    2.17e-07 4.50e+00 weight
    1.79e-06 4.65e+00 input[0]
    2.68e-06 3.70e+01 output
                      encoder.block.2.layer.1.DenseReluDense.wi_1 Linear
    8.08e-07 2.66e+01 weight
    1.79e-06 4.65e+00 input[0]
    1.27e-04 2.37e+02 output
                      encoder.block.2.layer.1.DenseReluDense.wo Linear
    1.01e-06 6.44e+00 weight
    0.00e+00 9.74e+03 input[0]
    3.18e-04 6.27e+04 output
                      encoder.block.2.layer.1.DenseReluDense T5DenseGatedGeluDense
    1.79e-06 4.65e+00 input[0]
    3.18e-04 6.27e+04 output
                      encoder.block.2.layer.1.dropout Dropout
    3.18e-04 6.27e+04 input[0]
    0.00e+00      inf output
    ```

    You can see here, that `T5DenseGatedGeluDense.forward` resulted in output activations, whose absolute max value
    was around 62.7K, which is very close to fp16's top limit of 64K. In the next frame we have `Dropout` which
    renormalizes the weights, after it zeroed some of the elements, which pushes the absolute max value to more than
    64K, and we get an overlow.

    As you can see it's the previous frames that we need to look into when the numbers start going into very large for
    fp16 numbers.

    The tracking is done in a forward hook, which gets invoked immediately after `forward` has completed.

    By default the last 21 frames are printed. You can change the default to adjust for your needs. For example :

    ```python
    debug_overflow = DebugUnderflowOverflow(model, max_frames_to_save=100)
    ```

        To validate that you have set up this debugging feature correctly, and you intend to use it in a training that may
        take hours to complete, first run it with normal tracing enabled for one of a few batches as explained in the next
        section.


        Mode 2. Specific batch absolute min/max tracing without detection

        The second work mode is per-batch tracing with the underflow/overflow detection feature turned off.

        Let's say you want to watch the absolute min and max values for all the ingredients of each `forward` call of a
    given batch, and only do that for batches 1 and 3. Then you instantiate this class as :

    ```python
    debug_overflow = DebugUnderflowOverflow(model, trace_batch_nums=[1, 3])
    ```

    And now full batches 1 and 3 will be traced using the same format as explained above. Batches are 0-indexed.

    This is helpful if you know that the program starts misbehaving after a certain batch number, so you can
    fast-forward right to that area.


    Early stopping:

    You can also specify the batch number after which to stop the training, with :

    ```python
    debug_overflow = DebugUnderflowOverflow(model, trace_batch_nums=[1, 3], abort_after_batch_num=3)
    ```

    This feature is mainly useful in the tracing mode, but you can use it for any mode.


    **Performance**:

    As this module measures absolute `min`/``max` of each weight of the model on every forward it'll slow the
    training down. Therefore remember to turn it off once the debugging needs have been met.

    Args:
        model (`nn.Module`):
            The model to debug.
        max_frames_to_save (`int`, *optional*, defaults to 21):
            How many frames back to record
        trace_batch_nums(`List[int]`, *optional*, defaults to `[]`):
            Which batch numbers to trace (turns detection off)
        abort_after_batch_num  (`int``, *optional*):
            Whether to abort after a certain batch number has finished
    """

    def __init__(
        self,
        model,
        max_frames_to_save=21,
        trace_batch_nums=[],
        abort_after_batch_num=None,
    ):
        self.model = model
        self.trace_batch_nums = trace_batch_nums
        self.abort_after_batch_num = abort_after_batch_num

        # keep a LIFO buffer of frames to dump as soon as inf/nan is encountered to give context to the problem emergence
        self.frames = collections.deque([], max_frames_to_save)
        self.frame = []
        self.batch_number = 0
        self.total_calls = 0
        self.detected_overflow = False
        self.prefix = "                 "

        self.analyse_model()

        self.register_forward_hook()

    def save_frame(self, frame=None):
        if frame is not None:
            self.expand_frame(frame)
        self.frames.append("\n".join(self.frame))
        self.frame = []  # start a new frame

    def expand_frame(self, line):
        self.frame.append(line)

    def trace_frames(self):
        print("\n".join(self.frames))
        self.frames = []

    def reset_saved_frames(self):
        self.frames = []

    def dump_saved_frames(self):
        print(f"\nDetected inf/nan during batch_number={self.batch_number}")
        print(f"Last {len(self.frames)} forward frames:")
        print(f"{'abs min':8} {'abs max':8} metadata")
        print("\n".join(self.frames))
        print("\n\n")
        self.frames = []

    def analyse_model(self):
        # extract the fully qualified module names, to be able to report at run time. e.g.:
        # encoder.block.2.layer.0.SelfAttention.o
        #
        # for shared weights only the first shared module name will be registered
        self.module_names = {m: name for name, m in self.model.named_modules()}
        # self.longest_module_name = max(len(v) for v in self.module_names.values())

    def analyse_variable(self, var, ctx):
        if torch.is_tensor(var):
            self.expand_frame(get_abs_min_max(var, ctx))
            if detect_overflow(var, ctx):
                self.detected_overflow = True
        elif var is None:
            self.expand_frame(f"{'None':>17} {ctx}")
        else:
            self.expand_frame(f"{'not a tensor':>17} {ctx}")

    def batch_start_frame(self):
        self.expand_frame(f"\n\n{self.prefix} *** Starting batch number={self.batch_number} ***")
        self.expand_frame(f"{'abs min':8} {'abs max':8} metadata")

    def batch_end_frame(self):
        self.expand_frame(f"{self.prefix} *** Finished batch number={self.batch_number-1} ***\n\n")

    def create_frame(self, module, input, output):
        self.expand_frame(f"{self.prefix} {self.module_names[module]} {module.__class__.__name__}")

        # params
        for name, p in module.named_parameters(recurse=False):
            self.analyse_variable(p, name)

        # inputs
        if isinstance(input, tuple):
            for i, x in enumerate(input):
                self.analyse_variable(x, f"input[{i}]")
        else:
            self.analyse_variable(input, "input")

        # outputs
        if isinstance(output, tuple):
            for i, x in enumerate(output):
                # possibly a tuple of tuples
                if isinstance(x, tuple):
                    for j, y in enumerate(x):
                        self.analyse_variable(y, f"output[{i}][{j}]")
                else:
                    self.analyse_variable(x, f"output[{i}]")
        else:
            self.analyse_variable(output, "output")

        self.save_frame()

    def register_forward_hook(self):
        self.model.apply(self._register_forward_hook)

    def _register_forward_hook(self, module):
        module.register_forward_hook(self.forward_hook)

    def forward_hook(self, module, input, output):
        # - input is a tuple of packed inputs (could be non-Tensors)
        # - output could be a Tensor or a tuple of Tensors and non-Tensors

        last_frame_of_batch = False

        trace_mode = True if self.batch_number in self.trace_batch_nums else False
        if trace_mode:
            self.reset_saved_frames()

        if self.total_calls == 0:
            self.batch_start_frame()
        self.total_calls += 1

        # count batch numbers - the very first forward hook of the batch will be called when the
        # batch completes - i.e. it gets called very last - we know this batch has finished
        if module == self.model:
            self.batch_number += 1
            last_frame_of_batch = True

        self.create_frame(module, input, output)

        # if last_frame_of_batch:
        #     self.batch_end_frame()

        if trace_mode:
            self.trace_frames()

        if last_frame_of_batch:
            self.batch_start_frame()

        if self.detected_overflow and not trace_mode:
            self.dump_saved_frames()

            # now we can abort, as it's pointless to continue running
            raise ValueError(
                "DebugUnderflowOverflow: inf/nan detected, aborting as there is no point running further. "
                "Please scroll up above this traceback to see the activation values prior to this event."
            )

        # abort after certain batch if requested to do so
        if self.abort_after_batch_num is not None and self.batch_number > self.abort_after_batch_num:
            raise ValueError(
                f"DebugUnderflowOverflow: aborting after {self.batch_number} batches due to `abort_after_batch_num={self.abort_after_batch_num}` arg"
            )


def get_abs_min_max(var, ctx):
    abs_var = var.abs()
    return f"{abs_var.min():8.2e} {abs_var.max():8.2e} {ctx}"


def detect_overflow(var, ctx):
    """Report whether the tensor contains any `nan` or `inf` entries.

    This is useful for detecting overflows/underflows and best to call right after the function that did some math that
    modified the tensor in question.

    This function contains a few other helper features that you can enable and tweak directly if you want to track
    various other things.

    Args:
        var: the tensor variable to check
        ctx: the message to print as a context

    Return:
        `True` if `inf` or `nan` was detected, `False` otherwise
    """
    detected = False
    if torch.isnan(var).any().item():
        detected = True
        print(f"{ctx} has nans")
    if torch.isinf(var).any().item():
        detected = True
        print(f"{ctx} has infs")

    # if needed to monitor large elements can enable the following
    if 0:  # and detected:
        n100 = var[torch.ge(var.abs(), 100)]
        if n100.numel() > 0:
            print(f"{ctx}:  n100={n100.numel()}")
        n1000 = var[torch.ge(var.abs(), 1000)]
        if n1000.numel() > 0:
            print(f"{ctx}: n1000={n1000.numel()}")
        n10000 = var[torch.ge(var.abs(), 10000)]
        if n10000.numel() > 0:
            print(f"{ctx}: n10000={n10000.numel()}")

    if 0:
        print(f"min={var.min():9.2e} max={var.max():9.2e}")

    if 0:
        print(f"min={var.min():9.2e} max={var.max():9.2e} var={var.var():9.2e} mean={var.mean():9.2e} ({ctx})")

    return detected



import logging
import os
import time
from pathlib import Path

import colorlog
import yaml
from torch.utils.tensorboard.writer import SummaryWriter

from audiozen.accelerate import is_rank_zero


class TensorboardLogger(SummaryWriter):
    def __init__(self, log_dir: str = "") -> None:
        super().__init__(log_dir=log_dir, max_queue=5, flush_secs=30)

    def log_config(self, config: dict) -> None:
        self.add_text(tag="Configuration", text_string=f"<pre>  \n{yaml.dump(config)}  \n</pre>", global_step=1)


def init_logging_logger(output_dir):
    """Initialize logging logger with `console` and `file` handlers.

    Note:
        We have two ways to log messages:
        - build-in `print` function.
        - `logging` module. It differs from the above two functions. It will log messages to a file and console. You
        may use it at all places in your code.

    Args:
        output_dir: Path to save log file.

    Examples:

        ```python
        # Call this function at the beginning of main file.
        init_logger(log_fpath="log_path")
        # Use this logger in other modules.
        import logging

        logger = logging.getLogger(__name__)
        logger.info("info message")
        ```
    """
    # Parse log_fpath
    output_dir = Path(output_dir) if isinstance(output_dir, str) else output_dir
    output_dir.mkdir(parents=True, exist_ok=True)

    # disable logging for libraries that use standard logging module
    logging.getLogger("matplotlib").setLevel(logging.WARNING)
    logging.getLogger("numba").setLevel(logging.ERROR)

    # Create root logger
    logger = logging.getLogger()

    # Remove all handlers associated with the root logger object
    if logger.hasHandlers():
        logger.handlers.clear()

    # Set the lowest level of root logger and controls logging via handlers' level
    logger.setLevel(logging.DEBUG)

    # Get log level from environment variable
    log_level = os.environ.get("LOG_LEVEL", "INFO").upper()

    # Create a console handler and set level to info
    console_handler = logging.StreamHandler()
    console_handler.setLevel(level=log_level)

    # Create a file handler and set the logger level to debug
    # Note: The file handler will log all messages to a file, regardless of the which GPU process it comes from.
    # Note: Here we use coarser time granularity to avoid the processes writing to different files at the same time.
    time_now = time.strftime("%Y_%m_%d--%H_%M")
    file_handler = logging.FileHandler(str(output_dir / f"{time_now}.log"))
    file_handler.setLevel(level=log_level)

    # Create formatters (file logger have more info)
    console_formatter = colorlog.ColoredFormatter(
        "%(log_color)s%(asctime)s: %(message)s",
        datefmt="%m-%d %H:%M:%S",
        log_colors={"DEBUG": "cyan", "INFO": "green", "WARNING": "yellow", "ERROR": "red", "CRITICAL": "bold_red"},
    )

    file_formatter = logging.Formatter(
        "%(asctime)s %(levelname)-8s [%(filename)s:%(lineno)d]: %(message)s", datefmt="%m-%d %H:%M:%S"
    )

    # Add formatter to ch
    console_handler.setFormatter(console_formatter)
    file_handler.setFormatter(file_formatter)

    # Add ch to logger
    logger.addHandler(console_handler)
    logger.addHandler(file_handler)

    if is_rank_zero():
        logger.info(f"Initialized logger with log file in {output_dir.as_posix()}.")
